/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:53
  filename: D:\Projects\Apgv5_0\Generator\CommandLine.c
   purpose: Reads and interprets the command line parameters
  
*********************************************************************/
#include "Generator.h"
#include <malloc.h>
#include <stdio.h>

static char* s_cpHelp1 = "?";
static char* s_cpHelp2 = "/?";
static char* s_cpHelp3 = "/h";
static char* s_cpIn = "/in";
static char* s_cpOut = "/out";
static char* s_cpLog = "/log";
static char* s_cpSestMax = "/sestmax";
static char* s_cpStartRule = "/startrule";
static char* s_cpNoRecursionAnalysis = "/nr";
static char* s_cpPpt = "/ppt";
static char* s_cpCpp = "/cpp";
static char* s_cpVerbose = "/dv";
static char* s_cpGrammar = "/dgrammar";
static char* s_cpErrors = "/derrors";
static char* s_cpWarnings = "/dwarnings";
static char* s_cpMetrics = "/dmetrics";
static char* s_cpState = "/dstate";
static char* s_cpTokens = "/dtokens";
static char* s_cpOpcodes = "/dopcodes";
static char* s_cpPPTs = "/dppts";
static char* s_cpRecursive = "/dr";
static char* s_cpRecursiveVerbose = "/drv";
static char* s_cpCharMap = "/dcharmap";
static char* s_cpWarranty = "/dwarranty";
static char* s_cpConditions = "/dconditions";

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Examines a command line option to see if it 1) matches
//              the specified option and 2) if it is followed by a file name.
//
// FUNCTION: static ulIsFileName
//
//      ARG: 1) char* cpArg
//              : the command line option to test
//
//      ARG: 2) char* cpOption
//              : the option to look for
//
//      ARG: 3) ulong* ulpError
//              : returns true on success
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsFileName(char* cpArg, char* cpOption, ulong* ulpError)
{
  ulong ulReturn = APG_TRUE;
  ulong ulLen = strlen(cpOption);
  
  // single-point-of-exit loop
  while(APG_TRUE)
  {
    if(0 != memcmp((void*)cpArg, cpOption, ulLen))
    {
      // input option not found
      ulReturn = APG_FALSE;
      break;
    }
    
    if(':' != cpArg[ulLen])
    {
      // format error
      printf("option error - \"%s\" option not followed by \":\"\n", cpArg);
      *ulpError = APG_TRUE;
      break;
    }
    
    ++ulLen;
    if(0 == cpArg[ulLen])
    {
      // format error
      printf("option error - \"%s\" file name is missing\n", cpArg);
      *ulpError = APG_TRUE;
      break;
    }
    
    // found good file name
    break;
  } // end of single-point-of-exit loop

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks for a specific command line option and extracts
//              an integer value for it.
//
// FUNCTION: static ulIsInteger
//
//      ARG: 1) char* cpArg
//              : the command line option to test
//
//      ARG: 2) char* cpOption
//              : the option to look for
//
//      ARG: 3) ulong* ulpInteger
//              : on successful return, the integer option value
//
//      ARG: 4) ulong* ulpError
//              : returns true on success
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsInteger(char* cpArg, char* cpOption, ulong* ulpInteger, ulong* ulpError)
{
  ulong ulReturn = APG_TRUE;
  ulong ulLen = strlen(cpOption);
  
  // single-point-of-exit loop
  while(APG_TRUE)
  {
    if(0 != memcmp((void*)cpArg, cpOption, ulLen))
    {
      // input option not found
      ulReturn = APG_FALSE;
      break;
    }
    
    if(':' != cpArg[ulLen])
    {
      // format error
      printf("option error - \"%s\" option not followed by \":\"\n", cpArg);
      *ulpError = APG_TRUE;
      break;
    }
    
    ++ulLen;
    if(0 == cpArg[ulLen])
    {
      // format error
      printf("option error - \"%s\" integer is missing\n", cpArg);
      *ulpError = APG_TRUE;
      break;
    }
    
    // found integer
    *ulpInteger = (ulong)atol(&cpArg[ulLen]);
    break;
  } // end of single-point-of-exit loop
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option to see if it is one of the
//              help option forms.
//
// FUNCTION: static ulIsHelp
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsHelp(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = APG_TRUE;

  // single-point-of-exit loop
  while(APG_TRUE)
  {
    if(0 == memcmp((void*)cpArg, s_cpHelp1, strlen(s_cpHelp1)))
    {
      spConfig->ulError = APG_TRUE;
      break;
    }
    
    if(0 == memcmp((void*)cpArg, s_cpHelp2, strlen(s_cpHelp2)))
    {
      spConfig->ulError = APG_TRUE;
      break;
    }

    if(0 == memcmp((void*)cpArg, s_cpHelp3, strlen(s_cpHelp3)))
    {
      spConfig->ulError = APG_TRUE;
      break;
    }
  
    ulReturn = APG_FALSE;
    break;
  } // end of single-point-of-exit loop

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option for the /in option and
//              extracts the file name.
//
// FUNCTION: static ulIsIn
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsIn(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = ulIsFileName(cpArg, s_cpIn, &spConfig->ulError);
  ulong ulLen = strlen(s_cpIn);

  if(ulReturn && !spConfig->ulError)
  {
    spConfig->ulInput = APG_TRUE;
    strcpy(spConfig->caInput, &cpArg[ulLen + 1]);
  }

  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option for the /out option and
//              extracts the file name.
//
// FUNCTION: static ulIsOut
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsOut(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = ulIsFileName(cpArg, s_cpOut, &spConfig->ulError);
  ulong ulLen = strlen(s_cpOut);
  
  if(ulReturn && !spConfig->ulError)
  {
    spConfig->ulParserName = APG_TRUE;
    strcpy(spConfig->caParserName, &cpArg[ulLen + 1]);
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option for the /log option and
//              extracts the file name.
//
// FUNCTION: static ulIsLog
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsLog(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = ulIsFileName(cpArg, s_cpLog, &spConfig->ulError);
  ulong ulLen = strlen(s_cpLog);
  
  if(ulReturn && !spConfig->ulError)
  {
    spConfig->ulLog = APG_TRUE;
    strcpy(spConfig->caLog, &cpArg[ulLen + 1]);
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option for the /sestmax option and
//              extracts the integer value.
//
// FUNCTION: static ulIsSestMax
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsSestMax(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = ulIsInteger(cpArg, s_cpSestMax, &spConfig->ulMaxTerminalNodes, &spConfig->ulError);
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks a command line option for the /startrule option and
//              extracts the integer value.
//
// FUNCTION: static ulIsStartRule
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulIsStartRule(char* cpArg, CM_CONFIG* spConfig)
{
  ulong ulReturn = ulIsInteger(cpArg, s_cpStartRule, &spConfig->ulStartRule, &spConfig->ulError);
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Checks for a specific option and returns its true/false value
//
// FUNCTION: static ulIsOption
//
//      ARG: 1) char* cpArg
//              : pointer to the command line option
//
//      ARG: 2) char* cpOption
//              : pointer to the option to look for
//
//      ARG: 3) ulong* ulpOption
//              : on return, true if the option was found, false otherwise
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static ulIsOption(char* cpArg, char* cpOption, ulong* ulpOption)
{
  ulong ulReturn = APG_FALSE;
  ulong ulLen = strlen(cpOption);
  if(0 == memcmp((void*)cpArg, cpOption, ulLen))
  {
    if(cpArg[ulLen] == 0)
    {
      *ulpOption = APG_TRUE;
      ulReturn = APG_TRUE;
    }
  }
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: parse the command line
//
// FUNCTION: ulong ulCommandLine
//
//      ARG: 1) CM_CONFIG* spConfig
//              : pointer to the command line configuration struct
//
//      ARG: 2) int argc
//              : the number of command line arguments
//
//      ARG: 3) char** argv
//              : pointer to a list of pointers to the command line arguments
//
//   RETURN: true on success
//           prints a help screen to the console on error or detection
//           of a help option
//
//////////////////////////////////////////////////////////////////////////
ulong ulCommandLine(CM_CONFIG* spConfig, int argc, char** argv)
{
  int i;
  char* cpArg;
  ulong ulReturn = APG_TRUE;
  ulong ulDisplayVerbose = APG_FALSE;
  ulong ulDisplayRuleAttrs = APG_FALSE;

  // set defaults - start rule = 0, all display is off
  memset((void*)spConfig, 0, sizeof(*spConfig));
  spConfig->ulMaxTerminalNodes = 1000000;

  for(i = 1; i < argc; ++i)
  {
    cpArg = argv[i];

    // help
    if(ulIsHelp(cpArg, spConfig)){break;}
    if(ulIsIn(cpArg, spConfig)){continue;}
    if(ulIsOut(cpArg, spConfig)){continue;}
    if(ulIsLog(cpArg, spConfig)){continue;}
    if(ulIsSestMax(cpArg, spConfig)){continue;}
    if(ulIsStartRule(cpArg, spConfig)){continue;}
    if(ulIsOption(cpArg, s_cpPpt, &spConfig->ulPPT)){continue;}
    if(ulIsOption(cpArg, s_cpNoRecursionAnalysis, &spConfig->ulNoRecursionAnalysis)){continue;}
    if(ulIsOption(cpArg, s_cpVerbose, &spConfig->ulDisplayVerbose)){continue;}
    if(ulIsOption(cpArg, s_cpGrammar, &spConfig->ulDisplayGrammar)){continue;}
    if(ulIsOption(cpArg, s_cpErrors, &spConfig->ulDisplayErrors)){continue;}
    if(ulIsOption(cpArg, s_cpWarnings, &spConfig->ulDisplayWarnings)){continue;}
    if(ulIsOption(cpArg, s_cpMetrics, &spConfig->ulDisplayMetrics)){continue;}
    if(ulIsOption(cpArg, s_cpRecursive, &spConfig->ulDisplayRecursion)){continue;}
    if(ulIsOption(cpArg, s_cpRecursiveVerbose, &spConfig->ulDisplayRuleAttrs)){continue;}
    if(ulIsOption(cpArg, s_cpTokens, &spConfig->ulDisplayTokens)){continue;}
    if(ulIsOption(cpArg, s_cpOpcodes, &spConfig->ulDisplayOpcodes)){continue;}
    if(ulIsOption(cpArg, s_cpPPTs, &spConfig->ulDisplayPPTs)){continue;}
    if(ulIsOption(cpArg, s_cpCpp, &spConfig->ulCpp)){continue;}
    if(ulIsOption(cpArg, s_cpCharMap, &spConfig->ulDisplayCharMap)){continue;}
    if(ulIsOption(cpArg, s_cpState, &spConfig->ulDisplayState)){continue;}

    // unknown option
    printf("option %d unrecognized: \"%s\"\n", i, cpArg);
    spConfig->ulError = APG_TRUE;
  }

  // input file name is required
  if(!spConfig->ulInput)
  {
    printf("/in:filename - input file name is required\n");
    spConfig->ulError = APG_TRUE;
  }

  if(spConfig->ulDisplayRuleAttrs)
  {
    // implies ulDisplayRecursion 
    spConfig->ulDisplayRecursion = APG_TRUE;
  }

  if(spConfig->ulDisplayVerbose)
  {
    // implies all other display options
    spConfig->ulDisplayGrammar = APG_TRUE;
    spConfig->ulDisplayErrors = APG_TRUE;
    spConfig->ulDisplayWarnings = APG_TRUE;
    spConfig->ulDisplayMetrics = APG_TRUE;
    spConfig->ulDisplayState = APG_TRUE;
    spConfig->ulDisplayRecursion = APG_TRUE;
  }
  else if(spConfig->ulDisplayErrors)
  {
    // implies
    spConfig->ulDisplayGrammar = APG_TRUE;
    spConfig->ulDisplayWarnings = APG_TRUE;
  }
  else if(spConfig->ulDisplayWarnings)
  {
    // implies
    spConfig->ulDisplayGrammar = APG_TRUE;
  }

  if(spConfig->ulError)
  {
    // command line errors
    printf("\n");
    printf("COMMAND LINE OPTIONS FOUND\n");
    for(i = 0; i < argc; ++i)
    {
      printf("%s \n", argv[i]);
    }
    printf("\n");
    printf("\n");

    // display the help screen
    printf("APG(version 5.0) Usage: Apg.exe options\n");
    printf("options: one or more of the following:\n");
    printf("\n");
    printf("help options\n");
    printf("?               - print this help screen\n");
    printf("/?              - print this help screen\n");
    printf("/h              - print this help screen\n");
    printf("\n");
    printf("file name options\n");
    printf("/in:filename    - input grammar file name (option required)\n");
    printf("/out:filebase   - generate parser to files filebase.h and filebase.c or filebase.cpp\n");
    printf("/log:filename   - log output to filename (default is console)\n");
    printf("\n");
    printf("integer value options\n");
    printf("/sestmax:int    - int is maximum number of SEST terminal nodes to allow (default = 1,000,000)\n");
    printf("/startrule:int  - int is start rule index (default = 0)\n");
    printf("\n");
    printf("true/false options\n");
    printf("/nr             - no recursion analysis - default is to do recursion analysis\n");
    printf("/ppt            - use Predictive Parsing Tables\n");
    printf("/cpp            - if /out:filebase is present, output a C++ parser component\n");
    printf("/dv             - verbose, implies display of all (v) options except\n");
    printf("/dgrammar       - (v) display input grammar file with one-based line numbers\n");
    printf("/dwarnings      - (v) display grammar syntax warnings (implies /dgrammar)\n");
    printf("/derrors        - (v) display grammar syntax errors (implies /dwarnings)\n");
    printf("/dmetrics       - (v) display grammar metrics\n");
    printf("/dstate         - (v) display parser states from both the lexical and token parsers\n");
    printf("/dr             - display recursion analysis\n");
    printf("/drv            - verbose recursion analysis, includes recursion attributes for individual rules\n");
    printf("/dcharmap       - display the map of used and unused alphabet characters (assumes ASCII character set)\n");
    printf("/dtokens        - (*) display grammar tokens\n");
    printf("/dopcodes       - (*) display grammar opcodes\n");
    printf("/dppts          - (*) display all PPT tables\n");

    printf("\n");
    printf("options must be lower case only\n");
    printf("option file names, if any, are case sensitive\n");
    printf("file name lengths, including extensions and null terminators, must be < 128\n");
    printf("default value of all true/false options are false\n");
    printf("(*) caveat: may generate extensive output\n");
    ulReturn = APG_FALSE;
  }

  return ulReturn;
}
