/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:55
  filename: D:\Projects\Apgv5_0\Generator\Generator.c
   purpose: Generates source and header files for APG parser components.
            Generates both C and C++ versions.
  
*********************************************************************/
#include "Generator.h"

// pre-defined strings
static char* GenTrans_CppFmt      = "%sulong %s(void* vpData, ulong ulState, ulong ulOffset, ulong ulLen)%s";
static char* cpHeaderPreBeg       = "//{{header.pre";
static char* cpHeaderPreEnd       = "//}}header.pre";
static char* cpHeaderPostBeg      = "//{{header.post";
static char* cpHeaderPostEnd      = "//}}header.post";
static char* cpClassPreBeg        = "//{{class.pre";
static char* cpClassPreEnd        = "//}}class.pre";
static char* cpClassPostBeg       = "//{{class.post";
static char* cpClassPostEnd       = "//}}class.post";
static char* cpDefinesPreBeg      = "//{{Include.pre";
static char* cpDefinesPreEnd      = "//}}Include.pre";
static char* cpDefinesPostBeg     = "//{{Include.post";
static char* cpDefinesPostEnd     = "//}}Include.post";
static char* cpCallBackPostBeg    = "//{{CallBack.body";
static char* cpCallBackPostEnd    = "//}}CallBack.body";
static char* cpCallBackBeg        = "//{{CallBack";
static char* cpCallBackEnd        = "//}}CallBack";
static char* cpGPLNotice =\
  "  APG - an ABNF Parser Generator\r\n"
  "  Copyright  (C) 2007 Coast to Coast Research, Inc.\r\n"
  " \r\n"
  "  This program is free software: you can redistribute it and/or modify\r\n"
  "  it under the terms of the GNU General Public License as published by\r\n"
  "  the Free Software Foundation, either version 2 of the License, or\r\n"
  "  (at your option) any later version.\r\n"
  "  \r\n"
  "  This program is distributed in the hope that it will be useful,\r\n"
  "  but WITHOUT ANY WARRANTY; without even the implied warranty of\r\n"
  "  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\r\n"
  "  GNU General Public License for more details.\r\n"
  "  \r\n"
  "  You should have received a copy of the GNU General Public License\r\n"
  "  along with this program.  If not, see\r\n"
  "  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>\r\n"
  "  or write to the Free Software Foundation, Inc.,\r\n"
  "  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.\r\n"
  "  \r\n";

#define APG_PAGE 4096
typedef void (*PFN_GENERATE)(GENCTX* spData, void* vpCtx, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName);
typedef struct 
{
  char* cpKey;
  ulong ulKeyLen;
  char* cpValue;
  ulong ulValueLen;
  ulong ulFound;
} DBENTRY;

static void vGenerate(GENCTX* spData, char* cpParserName);
static void vGenerateCode(GENCTX* spData, PFN_GENERATE pfnGenerator, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName);
static void vGenCallBackFunctionNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl);
static void vGenRuleIDNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl, char* cpParserName);
static void vGenerateHeader(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName);
static void vGenerateSource(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName);
static APG_EXEHDR* spGenerateExeData(GENCTX* spData);
static void vGenerateCppHeader(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpClassName, char* cpFileName);
static void vGenerateCppSource(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpClassName, char* cpFileName);
static void vGenCppRuleIDNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl, char* cpParserName);
static void vSaveUserCode(GENCTX* spData, void* vpSavedUserCodeCtx, char* cpFile);
static void vRestoreUserCode(GENCTX* spData, void* vpRestoredFileCtx, void* vpCodeFileCtx, void* vpSavedUserCodeCtx);
static int  iCompare(const void* vpLeft, const void* vpRight);
static void vWriteCharacterTable(void* vpCodeFileCtx, ulong* ulpTable, ulong ulTableSize);

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Generator component constructor
//
// FUNCTION: GENCTX* spGenCtor
//
//      ARG: 1) CM_CONFIG* spConfig
//              : pointer to a valid configuration struct
//
//   RETURN: pointer to the generator component context.
//
//////////////////////////////////////////////////////////////////////////
GENCTX* spGenCtor(CM_CONFIG* spConfig)
{
  GENCTX* spReturn = NULL;
  GENCTX* spData = NULL;
  void* vpMemCtx;
  char* cpMsg = "spGenCtor: can't allocate memory";

  // get a private memory component
  vpMemCtx = vpMemCtor();
  if(!vpMemCtx)
  {
    vPrintLoc(cpMsg, __LINE__, __FILE__);
    goto APG_EXIT;
  }


  // allocate & clear the data
  spData = (GENCTX*)vpMemAlloc(vpMemCtx, sizeof(GENCTX));
  if(!spData)
  {
    vPrintLoc(cpMsg, __LINE__, __FILE__);
    goto APG_EXIT;
  }

  memset((void*)spData, 0, sizeof(*spData));
  spData->vpMemCtx = vpMemCtx;
  spData->sConfig = *spConfig;

  spData->vpLinesCtx = vpVecCtor(spData->vpMemCtx, sizeof(APG_BSTR), 0);
  if(!spData->vpLinesCtx)
  {
    vPrintLoc(cpMsg, __LINE__, __FILE__);
    goto APG_EXIT;
  }
  
  // set the PPT flag
  spData->ulPPT = spConfig->ulPPT;
  spData->ulStartRule = spConfig->ulStartRule;
  spData->vpValidate = (void*)spData;
  
  // success
  spReturn = spData;
APG_EXIT:
  if(!spReturn)
  {
    if(vpMemCtx)
    {
      // release all memory allocations
      spData->vpValidate = NULL;
      vMemDtor(vpMemCtx);
    }
  }
  return spReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Generator component destructor.
//
// FUNCTION: void vGenDtor
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vGenDtor(GENCTX* spData)
{
  // ignore null pointers
  if(spData)
  {
    DASSERT(spData->vpValidate == (void*)spData);

    // delete any remaining allocations, including the memory component
    vMemDtor(spData->vpMemCtx);
  }
}

typedef struct 
{
  ulong ulOpOffset;
  ulong ulLastIndex;
} INDEX_STACK;

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Generate a parser component source and header file.
//
// FUNCTION: ulong ulGenOutput
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) char* cpParserName
//              : base name of the source and header files
//                file names are base.h and base.c or base.cpp
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong   ulGenOutput(GENCTX* spData, char* cpParserName)
{
  ulong ulReturn = APG_FALSE;
  ulong ulTest;

  // generate a parser header & source files
  vGenerate(spData, cpParserName);
  
  ulTest = ulErrCount(spData->vpMemCtx);
  if(ulTest == 0)
  {
    ulReturn = APG_TRUE;
  }

  return ulReturn;
}

// print errors and warnings
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Print all errors and warnings if requested in the configuration.
//
// FUNCTION: void vGenErrors
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void    vGenErrors(GENCTX* spData)
{
  ulong ulDisplayErrors;
  ulong ulCount;
  ulong ulTotal;
  ulong ulIndex;
  ulong ulLine;
  char  caLine[APG_ERROR_BUFLEN];
  char  caMsg[APG_ERROR_BUFLEN];
  
  ulCount = ulGrammarErrorsCount(spData->vpErrorsCtx);
  ulDisplayErrors = ulCount || spData->sConfig.ulDisplayErrors;

  if(ulDisplayErrors)
  {
    printf("\n");
    printf("*** GRAMMAR ERRORS[%lu]\n", ulCount);
    if(ulCount)
    {
      vGrammarErrorsInit(spData->vpErrorsCtx);
      for(ulIndex = 0; ulIndex < ulCount; ++ulIndex)
      {
        vGrammarErrorsNext(spData->vpErrorsCtx, &ulLine, caLine, caMsg);
        printf("\n");
        printf("[%lu] %s\n", ulLine, caLine);
        printf("[%lu] %s\n", ulIndex, caMsg);
      }
    }
    else
    {
      printf("<none>\n");
    }
  }
  
  if(spData->sConfig.ulDisplayWarnings)
  {
    ulCount = ulGrammarErrorsCount(spData->vpWarningsCtx);
    printf("\n");
    printf("*** GRAMMAR WARNINGS[%lu]\n", ulCount);
    if(ulCount)
    {
      vGrammarErrorsInit(spData->vpWarningsCtx);
      for(ulIndex = 0; ulIndex < ulCount; ++ulIndex)
      {
        vGrammarErrorsNext(spData->vpWarningsCtx, &ulLine, caLine, caMsg);
        printf("\n");
        printf("[%lu] %s\n", ulLine, caLine);
        printf("[%lu] %s\n", ulIndex, caMsg);
      }
    }
    else
    {
      printf("<none>\n");
    }
  }

  ulTotal = ulErrTotal(spData->vpMemCtx);
  ulCount = ulErrCount(spData->vpMemCtx);
  if(ulCount)
  {
    // display system errors
    printf("\n");
    printf("*** SYSTEM ERRORS\n");
    printf("***      total [%lu]\n", ulTotal);
    printf("***   reported [%lu]\n", ulCount);
    printf("*** unreported [%lu]\n", (ulTotal - ulCount));
    vPrintErrors(spData->vpMemCtx);
  }

}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Generate parser component source and header files.
//
// FUNCTION: void vGenerate
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) char* cpParserName
//              : base name to use for the files
//                generates base.h and base.c or base.cpp
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void  vGenerate(GENCTX* spData, char* cpParserName)
{
  ulong ulNameLen;
  char* cpFileName;
  ulong ulRuleCount;
  char* cpStrTbl;
  APG_RULES* spRules;
  APG_EXEHDR* spExe = NULL;

  GENASSERT(spData);
  GENASSERT(cpParserName);

  // generate header name
  ulNameLen = strlen(cpParserName);
  cpFileName = (char*)vpMemAlloc(spData->vpMemCtx, ulNameLen + 5);
  GENASSERT(cpFileName);
  strcpy(cpFileName, cpParserName);
  cpFileName[ulNameLen] = '.';
  cpFileName[ulNameLen + 1] = 'h';
  cpFileName[ulNameLen + 2] = 0;

  // create vectors for call back and rule ID names
  spData->vpTransNamesCtx = vpVecCtor(spData->vpMemCtx, sizeof(char), spData->ulRuleCount * 10);
  GENASSERT(spData->vpTransNamesCtx);
  spData->vpIDNamesCtx = vpVecCtor(spData->vpMemCtx, sizeof(char), spData->ulRuleCount * 10);
  GENASSERT(spData->vpIDNamesCtx);

  // generate a list of call back function names
  cpStrTbl = (char*)vpVecFront(spData->vpStrTblCtx);
  ulRuleCount = spData->ulRuleCount;
  spRules = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  vGenCallBackFunctionNames(spData->vpTransNamesCtx, spRules, ulRuleCount, cpStrTbl);

  // generate the executable data
  spExe = spGenerateExeData(spData);
  GENASSERT(spExe);

  if(spData->sConfig.ulCpp)
  {
    char* cpClassName = (char*)vpMemAlloc(spData->vpMemCtx, strlen(cpParserName) + 5);
    GENASSERT(cpClassName);
    strcpy(cpClassName, "Apg");
    strcat(cpClassName, cpParserName);

    // generate a list of rule name IDs
    vGenCppRuleIDNames(spData->vpIDNamesCtx, spRules, ulRuleCount, cpStrTbl, cpClassName);
    
    // generate header code
    vGenerateCode(spData, vGenerateCppHeader, spExe, cpClassName, cpFileName);
    
    // generate source name
    cpFileName[ulNameLen + 1] = 'c';
    cpFileName[ulNameLen + 2] = 'p';
    cpFileName[ulNameLen + 3] = 'p';
    cpFileName[ulNameLen + 4] = 0;
    
    // generate source code
    vGenerateCode(spData, vGenerateCppSource, spExe, cpClassName, cpFileName);
  }
  else
  {
    // generate a list of rule name IDs
    vGenRuleIDNames(spData->vpIDNamesCtx, spRules, ulRuleCount, cpStrTbl, cpParserName);
    
    // generate header code
    vGenerateCode(spData, vGenerateHeader, spExe, cpParserName, cpFileName);
    
    // generate source name
    cpFileName[ulNameLen + 1] = 'c';
    
    // generate source code
    vGenerateCode(spData, vGenerateSource, spExe, cpParserName, cpFileName);
  }

APG_EXIT:
  // release memory resources
  vMemFree(spData->vpMemCtx, (void*)spExe);
  vMemFree(spData->vpMemCtx, (void*)cpFileName);
  vVecDtor(spData->vpTransNamesCtx);
  spData->vpTransNamesCtx = NULL;
  vVecDtor(spData->vpIDNamesCtx);
  spData->vpIDNamesCtx = NULL;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: 1) Saves user code, if any
//              2) Generates the output file (source or header)
//              3) Restores any user code
//              4) Writes the file to disk
//
// FUNCTION: static void vGenerateCode
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) PFN_GENERATE pfnGenerator
//              : pointer to either the source or header file generator
//
//      ARG: 3) APG_EXEHDR* spExe
//              : pointer to a structure to recieve the executable file data.
//
//      ARG: 4) char* cpParserName
//              : pointer to the parser component name
//
//      ARG: 5) char* cpFileName
//              : pointer to the disk file name
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenerateCode(GENCTX* spData, PFN_GENERATE pfnGenerator, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName)
{
  FILE* spFile = NULL;
  void* vpSavedUserCodeCtx = NULL;
  void* vpCodeFileCtx = NULL;
  void* vpRestoredFileCtx = NULL;
  char* cpOldFile;
  ulong ulOldFileSize;
  char* cpNewFile;
  ulong ulNewFileSize;
  ulong ulTest;
  ulong ulChkPt;

  // memory check point
  ulChkPt = ulMemCheckPoint(spData->vpMemCtx);

  // create required vectors
  vpCodeFileCtx = vpVecCtor(spData->vpMemCtx, 1, APG_PAGE);
  vpSavedUserCodeCtx = vpVecCtor(spData->vpMemCtx, sizeof(DBENTRY), 100);
  vpRestoredFileCtx = vpVecCtor(spData->vpMemCtx, 1, APG_PAGE);

  // check for old file
  ulOldFileSize = ulGetFileSize(cpFileName);
  if(ulOldFileSize)
  {
    // read old file & add a null term
    cpOldFile = (char*)vpMemAlloc(spData->vpMemCtx, ulOldFileSize + 1);
    GENASSERT(cpOldFile);
    cpOldFile[ulOldFileSize] = 0;
    ulTest = ulGetFile(cpFileName, (void*)cpOldFile);
    GENASSERT(ulTest == ulOldFileSize);

    // save the user code if any
    vSaveUserCode(spData, vpSavedUserCodeCtx, cpOldFile);
  }


  // generate the new code file
  pfnGenerator(spData, vpCodeFileCtx, spExe, cpParserName, cpFileName);

  if(ulOldFileSize)
  {
    if(vpSavedUserCodeCtx)
    {
      // restore the user code
      vRestoreUserCode(spData, vpRestoredFileCtx, vpCodeFileCtx, vpSavedUserCodeCtx);
    }

    // delete the old file
    remove(cpFileName);
    vpCodeFileCtx = vpRestoredFileCtx;
  }

  // remove the null term & write the new file
  cpNewFile = (char*)vpVecFront(vpCodeFileCtx);
  ulNewFileSize = ulVecSize(vpCodeFileCtx);
  ulTest = ulWriteFile(cpFileName, (void*)cpNewFile, ulNewFileSize);
  if(ulTest != ulNewFileSize)
  {
    printf("\n");
    printf("*** Write new file \"%s\" failed. May be read-only. ***\n", cpFileName);
    GENASSERT(APG_FALSE);
  }
  
APG_EXIT:
  // clean up all memory allocations
  vMemFreeToCheckPoint(spData->vpMemCtx, ulChkPt);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Creates a C language header file in a memory buffer (a byte vector)
//
// FUNCTION: static void vGenerateHeader
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpCodeFileCtx
//              : pinter to a byte-element vector component context 
//
//      ARG: 3) APG_EXEHDR* spExe
//              : pointer to an executable data struct
//
//      ARG: 4) char* cpParserName
//              : the name of the parser
//
//      ARG: 5) char* cpFileName
//              : the name of the output file
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenerateHeader(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName)
{
  char*           cpTransNames = NULL;
  char*           cpIDNames = NULL;
  ulong           ulStringCount;
  uchar*          ucpStrTbl;
  ulong           ulRuleCount;
  APG_RULES*      spRuleList;
  ulong           ul;
  char*           cpID;
  struct tm       *newTime;
  char*           cpAsciiTime;
  time_t          szClock;
  char            caBuf[APG_PAGE];
  ulong           ulBufLen;

  // set up control variables
  ulStringCount = ulVecSize(spData->vpStrTblCtx);
  ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  ulRuleCount = spData->ulRuleCount;
  spRuleList = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  cpTransNames = (char*)vpVecFront(spData->vpTransNamesCtx);
  cpIDNames = (char*)vpVecFront(spData->vpIDNamesCtx);
  
  //////////////////////////////////////////////////////////////////////////
  // HEADER
  //////////////////////////////////////////////////////////////////////////
  time( &szClock );
  newTime = localtime( &szClock );
  cpAsciiTime = asctime(newTime);
  ulBufLen = (ulong)sprintf(caBuf, "/********************************************************************\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // insert the GPL copyright notice
  ulBufLen = strlen(cpGPLNotice);
  vpVecPushn(vpCodeFileCtx, (void*)cpGPLNotice, ulBufLen);
  
  // NOTE: cpAsciiTime ends in \n only
  ulBufLen = (ulong)sprintf(caBuf, "	 created: ");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)strlen(cpAsciiTime);
  vpVecPushn(vpCodeFileCtx, (void*)cpAsciiTime, ulBufLen - 1);
  vpVecPushn(vpCodeFileCtx, (void*)"\r\n", 2);
  
  ulBufLen = (ulong)sprintf(caBuf, "	filename: %s\r\n", cpFileName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	  author: Lowell Thomas\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          lowell@coasttocoastresearch.com\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          http://www.coasttocoastresearch.com\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	 purpose: Header file for parser %s.\r\n\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*********************************************************************/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#ifndef %s_h\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#define %s_h\r\n\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // user pre-header markers
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpHeaderPreBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpHeaderPreEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // generate rule name IDs
  ulBufLen = (ulong)sprintf(caBuf, "// rule name IDs\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  cpID = cpIDNames;
  for(ul = 0; ul < ulRuleCount; ul++)
  {
    if(*cpID == 0){break;}
    ulBufLen = (ulong)sprintf(caBuf, "#define %s %lu\r\n", cpID, ul);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    cpID += strlen(cpID) + 1;
  }

  // generate the uniquely named pointer to the parser context
  ulBufLen = (ulong)sprintf(caBuf, "\r\nextern void* vp%sInit;\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // user post-header markers
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s\r\n", cpHeaderPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpHeaderPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // complete the file
  ulBufLen = (ulong)sprintf(caBuf, "#endif // %s_h\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Generates relocatable parser executable data.
//
// FUNCTION: static APG_EXEHDR* spGenerateExeData
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: pointer to the executable data struct, NULL on failure
//
//////////////////////////////////////////////////////////////////////////
static APG_EXEHDR* spGenerateExeData(GENCTX* spData)
{
  APG_EXEHDR*       spReturn = NULL;
  ulong             ulRuleCount = spData->ulRuleCount;
  APG_RULES*        spRule;
  void*             vpSrc;
  void*             vpDst;
  ulong             ulBytes, ulWords; // word = ulong = 4 bytes
  ulong             ul;
  ulong             ulTransCount;
  ulong             ulWriteUlongs;
  uchar*            ucpBuffer;
  GRAMMAR_METRICS*  spGMetrics = &spData->sGrammarMetrics;
  APG_EXEHDR*       spExe = NULL;
  APG_EXEHDR        sExe;

  // create an executable data header
  memset((void*)&sExe, 0, sizeof(sExe));

  // header info
  sExe.ulRuleCount = spData->ulRuleCount;
  sExe.ulCharSetSize = spData->ulCharSetSize;
  sExe.ulEOFChar = spData->ulEOFChar;
  sExe.ulMaxChar = spData->ulMaxChar;
  sExe.ulOpcodeCount = spData->ulOpcodeCount;;

  // header sections
  sExe.sHeader.ulOffset = 0;
  sExe.sHeader.ulLen = sizeof(sExe);

  // string table
  sExe.sStrTbl.ulOffset = sExe.sHeader.ulOffset + sExe.sHeader.ulLen;
  ulBytes = ulVecElementSize(spData->vpStrTblCtx) * ulVecSize(spData->vpStrTblCtx);
  ulWords = ulCeiling4(ulBytes);
  sExe.sStrTbl.ulLen = sizeof(ulong) * ulWords;

  // opcodes
  sExe.sOpcodes.ulOffset = sExe.sStrTbl.ulOffset + sExe.sStrTbl.ulLen;
  ulBytes = ulVecElementSize(spData->vpOpcodeCtx) * ulVecSize(spData->vpOpcodeCtx);
  ulWords = ulCeiling4(ulBytes);
  sExe.sOpcodes.ulLen = sizeof(ulong) * ulWords;

  // rule list
  sExe.sRuleList.ulOffset = sExe.sOpcodes.ulOffset + sExe.sOpcodes.ulLen;
  ulBytes = ulVecElementSize(spData->vpRuleCtx) * ulVecSize(spData->vpRuleCtx);
  ulWords = ulCeiling4(ulBytes);
  sExe.sRuleList.ulLen = sizeof(ulong) * ulWords;
  

  // PPT info
  if(spData->ulPPT)
  {
    sExe.ulPPT = APG_TRUE;
    sExe.sPPT.ulOffset = sExe.sRuleList.ulOffset + sExe.sRuleList.ulLen;
    ulBytes = ulVecElementSize(spData->vpPPTCtx) * ulVecSize(spData->vpPPTCtx);
    ulWords = ulCeiling4(ulBytes);
    sExe.sPPT.ulLen = sizeof(ulong) * ulWords;
  }

  // compute total buffer size
  sExe.ulLength = 0;
  sExe.ulLength += sExe.sHeader.ulLen;
  sExe.ulLength += sExe.sStrTbl.ulLen;
  sExe.ulLength += sExe.sOpcodes.ulLen;
  sExe.ulLength += sExe.sRuleList.ulLen;
  sExe.ulLength += sExe.sPPT.ulLen;
  ulWriteUlongs = ulCeiling4(sExe.ulLength);
  DASSERT(sizeof(ulong) * ulWriteUlongs == sExe.ulLength);

  // get the buffer and zero it out
  spExe = (APG_EXEHDR*)vpMemAlloc(spData->vpMemCtx, ulWriteUlongs * sizeof(ulong));
  GENASSERT(spExe);
  memset((void*)spExe, 0, ulWriteUlongs * sizeof(ulong)); 

  // copy data to the buffer
  // NOTE: be careful to only write the actual number of data bytes
  //       the buffer space for each has been rounded up, when necessary,
  //       to an integral number of 4-byte words

  // copy the header itself
  ucpBuffer = (uchar*)spExe;
  vpDst = (void*)ucpBuffer;
  vpSrc = (void*)&sExe;
  ulBytes = sExe.sHeader.ulLen;
  memcpy(vpDst, vpSrc, ulBytes);
  ucpBuffer += spExe->sHeader.ulLen;

  // copy the string table
  vpDst = (void*)ucpBuffer;
  vpSrc = vpVecFront(spData->vpStrTblCtx);
  ulBytes = ulVecElementSize(spData->vpStrTblCtx) * ulVecSize(spData->vpStrTblCtx);
  memcpy(vpDst, vpSrc, ulBytes);
  ucpBuffer += spExe->sStrTbl.ulLen;

  // copy the production opcodes
  vpDst = (void*)ucpBuffer;
  vpSrc = vpVecFront(spData->vpOpcodeCtx);
  ulBytes = ulVecElementSize(spData->vpOpcodeCtx) * ulVecSize(spData->vpOpcodeCtx);
  memcpy(vpDst, vpSrc, ulBytes);
  ucpBuffer += spExe->sOpcodes.ulLen;

  // re-index the call back function pointers in the rule to reflect only those kept
  spRule = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  ulTransCount = 0;
  for(ul = 0; ul < ulRuleCount; ul++, ++spRule)
  {
    if(spRule->usSyntax || spRule->usSemantics)
    {
      spRule->pfnCallBack = (PFN_CALLBACK)ulTransCount;
      ++ulTransCount;
    }
    else
    {
      spRule->pfnCallBack = (PFN_CALLBACK)UNDEFINED;
    }

    if(!sExe.ulPPT)
    {
      spRule->ucpPPT = NULL;
    }
  }

  // copy the production rule list
  vpDst = (void*)ucpBuffer;
  vpSrc = vpVecFront(spData->vpRuleCtx);
  ulBytes = ulVecElementSize(spData->vpRuleCtx) * ulVecSize(spData->vpRuleCtx);
  memcpy(vpDst, vpSrc, ulBytes);
  ucpBuffer += spExe->sRuleList.ulLen;

  if(spData->ulPPT)
  {
    // copy the PPTs
    vpDst = (void*)ucpBuffer;
    vpSrc = vpVecFront(spData->vpPPTCtx);
    ulBytes = ulVecElementSize(spData->vpPPTCtx) * ulVecSize(spData->vpPPTCtx);
    memcpy(vpDst, vpSrc, ulBytes);
    ucpBuffer += spExe->sRuleList.ulLen;
  }

  // add a hash of the data
  spExe->ulCrc32 = 0;
  ul = ulCRC32((uchar*)spExe, spExe->ulLength);
  spExe->ulCrc32 = ul;

  // success
  spReturn = spExe;

APG_EXIT:
  return spReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Creates a C language source file in a memory buffer (a byte vector)
//
// FUNCTION: static void vGenerateSource
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpCodeFileCtx
//              : pinter to a byte-element vector component context 
//
//      ARG: 3) APG_EXEHDR* spExe
//              : pointer to an executable data struct
//
//      ARG: 4) char* cpParserName
//              : the name of the parser
//
//      ARG: 5) char* cpFileName
//              : the name of the output file
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenerateSource(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpParserName, char* cpFileName)
{
  uchar*            ucpStrTbl;
  char*             cpRuleName;
  ulong             ulRuleCount = spData->ulRuleCount;
  APG_RULES*        spRule;
  ulong             ul;
  ulong             ulTransCount;
  char*             cpTrans;
  struct tm*        newTime;
  char*             cpAsciiTime;
  time_t            szClock;
  char              caBuf[APG_PAGE];
  ulong             ulBufLen;
  ulong             ulWriteUlongs;
  GRAMMAR_METRICS*  spGMetrics = &spData->sGrammarMetrics;


  // write the intro
  time( &szClock );
  newTime = localtime( &szClock );
  cpAsciiTime = asctime(newTime);
  ulBufLen = (ulong)sprintf(caBuf, "/********************************************************************\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // insert the GPL copyright notice
  ulBufLen = strlen(cpGPLNotice);
  vpVecPushn(vpCodeFileCtx, (void*)cpGPLNotice, ulBufLen);
  
  // NOTE: cpAsciiTime ends in \n only
  ulBufLen = (ulong)sprintf(caBuf, "	 created: ");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)strlen(cpAsciiTime);
  vpVecPushn(vpCodeFileCtx, (void*)cpAsciiTime, ulBufLen - 1);
  vpVecPushn(vpCodeFileCtx, (void*)"\r\n", 2);
  
  ulBufLen = (ulong)sprintf(caBuf, "	filename: %s\r\n", cpFileName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	  author: Lowell Thomas\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          lowell@coasttocoastresearch.com\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          http://www.coasttocoastresearch.com\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	 purpose: Implementation of parser component %s.\r\n\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*********************************************************************/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#include \"Apg.h\"\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpDefinesPreBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpDefinesPreEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#include \"%s.h\"\r\n\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpDefinesPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpDefinesPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // write general info
  ulBufLen = (ulong)sprintf(caBuf, "//  GENERAL INFORMATION:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  {
    ulong ulTotal = spExe->ulLength + spExe->ulRuleCount * sizeof(PFN_CALLBACK);
    double fKB = (double)ulTotal / 1024.0;
    double fMB = (double)ulTotal / (1024.0 * 1024.0);
    ulBufLen = (ulong)sprintf(caBuf, "//       static data = %lu (bytes)\r\n", ulTotal);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    ulBufLen = (ulong)sprintf(caBuf, "//                   = %.2E (KB)\r\n", fKB);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    ulBufLen = (ulong)sprintf(caBuf, "//                   = %.2E (MB)\r\n", fMB);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  ulBufLen = (ulong)sprintf(caBuf, "//   character count = %lu (grammar alphabet character set)\r\n", spData->ulCharSetSize);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "// unused characters = %lu (grammar alphabet character set)\r\n", spData->ulCharSetUnused);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//        rule count = %lu\r\n", spData->ulRuleCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//      opcode count = %lu\r\n", spData->ulOpcodeCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         ALT count = %lu(%lu) (alternations)\r\n", spGMetrics->ulALT, spGMetrics->ulAlternates);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         CAT count = %lu(%lu) (concatenations)\r\n", spGMetrics->ulCAT, spGMetrics->ulConcats);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         RNM count = %lu\r\n", spGMetrics->ulRNM);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         REP count = %lu\r\n", spGMetrics->ulREP);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         PRD count = %lu\r\n", spGMetrics->ulPRD);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TRG count = %lu\r\n", spGMetrics->ulTRG);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TBS count = %lu\r\n", spGMetrics->ulTBS);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TLS count = %lu\r\n", spGMetrics->ulTLS);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//      string table = %lu (bytes)\r\n", spExe->sStrTbl.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//           opcodes = %lu (bytes)\r\n", spExe->sOpcodes.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         rule list = %lu (bytes)\r\n", spExe->sRuleList.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  if(spData->ulPPT)
  {
    ulBufLen = (ulong)sprintf(caBuf, "//         PPT bytes = %lu\r\n", spExe->sPPT.ulLen);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  else
  {
    ulBufLen = (ulong)sprintf(caBuf, "//              PPTs = <not used>\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }

  // write the call back function declarations
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "// CALL BACK FUNCTION DECLARATIONS\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
  ulTransCount = 0;
  for(ul = 0; ul < ulRuleCount; ul++)
  {
    if(*cpTrans == 0){break;}
    ulBufLen = (ulong)sprintf(caBuf, GenTrans_CppFmt, "static ", cpTrans, ";\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    cpTrans += strlen(cpTrans) + 1;
    ++ulTransCount;
  }

  // write the call back function pointers
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "// CALL BACK FUNCTION POINTERS\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "static ulong g_ulCallbacks = %lu;\r\n", ulTransCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  if(ulTransCount > 0)
  {
    ulBufLen = (ulong)sprintf(caBuf, "static PFN_CALLBACK g_pfnCallbacks[%d] = \r\n{\r\n", ulTransCount);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    
    cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
    for(ul = 0; ul < ulRuleCount; ul++)
    {
      if(*cpTrans == 0){break;}
      ulBufLen = (ulong)sprintf(caBuf, "  %s,\r\n", cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      cpTrans += strlen(cpTrans) + 1;
    }
    
    // finish call back function pointers
    ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  else
  {
    // handle case of zero call backs
    ulBufLen = (ulong)sprintf(caBuf, "static PFN_CALLBACK g_pfnCallbacks[1];\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  
  // executable data initialization
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// EXECUTABLE DATA\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulWriteUlongs = spExe->ulLength / sizeof(ulong);
  ulBufLen = (ulong)sprintf(caBuf, "static ulong g_ulaExeData[%d] = \r\n{\r\n", ulWriteUlongs);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // output the data
  vWriteCharacterTable(vpCodeFileCtx, (ulong*)spExe, ulWriteUlongs);
  
  // finish executable data initialization
  ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // generate the unique parser named context pointer
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// UNIQUELY NAMED PARSER INITIALIZATION DATA\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "static APG_PARSER_INIT g_sParserInit = \r\n{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  ulBufLen = (ulong)sprintf(caBuf, "  0,\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  (void*)g_pfnCallbacks,\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  (void*)g_ulaExeData\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // finish unique parser named context pointer
  ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "void* vp%sInit = (void*)&g_sParserInit;\r\n", cpParserName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // switch template for call back functions
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// switch template\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "/*\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  switch(ulState)\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  {\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_PRE:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_NOMATCH:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_EMPTY:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_MATCH:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SEM_PRE:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SEM_POST:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  }\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s\r\n", cpCallBackPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpCallBackPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // generate call back functions
  spRule = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
  ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  for(ul = 0; ul < ulRuleCount; ul++, spRule++)
  {
    if(spRule->usSyntax || spRule->usSemantics)
    {
      // add a call back function
      cpRuleName = (char*)ucpStrTbl + (ulong)spRule->cpRuleName;
      ulBufLen = (ulong)sprintf(caBuf, "\r\n// call back function for rule \"%s\"\r\n", cpRuleName);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, GenTrans_CppFmt, "static ", cpTrans, "\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf,  "{\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "  ulong ulReturn = PFN_OK;\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "%s.%s\r\n", cpCallBackBeg, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "%s.%s\r\n", cpCallBackEnd, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "  return ulReturn;\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      cpTrans += strlen(cpTrans) + 1;
    }

  }
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Creates a C++ language header file in a memory buffer (a byte vector)
//
// FUNCTION: static void vGenerateCppHeader
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpCodeFileCtx
//              : pinter to a byte-element vector component context 
//
//      ARG: 3) APG_EXEHDR* spExe
//              : pointer to an executable data struct
//
//      ARG: 4) char* cpParserName
//              : the name of the parser
//
//      ARG: 5) char* cpFileName
//              : the name of the output file
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenerateCppHeader(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpClassName, char* cpFileName)
{
  char*           cpTransNames = NULL;
  char*           cpIDNames = NULL;
  ulong           ulStringCount;
  uchar*          ucpStrTbl;
  ulong           ulRuleCount;
  APG_RULES*      spRuleList;
  ulong           ul;
  char*           cpID;
  struct tm       *newTime;
  char*           cpAsciiTime;
  time_t          szClock;
  char            caBuf[APG_PAGE];
  ulong           ulBufLen;

  // set up control variables
  ulStringCount = ulVecSize(spData->vpStrTblCtx);
  ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  ulRuleCount = spData->ulRuleCount;
  spRuleList = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  cpTransNames = (char*)vpVecFront(spData->vpTransNamesCtx);
  cpIDNames = (char*)vpVecFront(spData->vpIDNamesCtx);
  
  //////////////////////////////////////////////////////////////////////////
  // HEADER
  //////////////////////////////////////////////////////////////////////////
  time( &szClock );
  newTime = localtime( &szClock );
  cpAsciiTime = asctime(newTime);
  ulBufLen = (ulong)sprintf(caBuf, "/********************************************************************\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // insert the GPL copyright notice
  ulBufLen = strlen(cpGPLNotice);
  vpVecPushn(vpCodeFileCtx, (void*)cpGPLNotice, ulBufLen);
  
  // NOTE: cpAsciiTime ends in \n only
  ulBufLen = (ulong)sprintf(caBuf, "	 created: ");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)strlen(cpAsciiTime);
  vpVecPushn(vpCodeFileCtx, (void*)cpAsciiTime, ulBufLen - 1);
  vpVecPushn(vpCodeFileCtx, (void*)"\r\n", 2);
  
  ulBufLen = (ulong)sprintf(caBuf, "	filename: %s\r\n", cpFileName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	  author: Lowell Thomas\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          lowell@coasttocoastresearch.com\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          http://www.coasttocoastresearch.com\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	 purpose: Header file for parser class %s.\r\n\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*********************************************************************/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#ifndef %s_h\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#define %s_h\r\n\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // user pre-header markers
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpHeaderPreBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpHeaderPreEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // start the class
  ulBufLen = (ulong)sprintf(caBuf, "class %s\r\n{\r\npublic:\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s\r\n", cpClassPreBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpClassPreEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n  //RULE NAME IDs\r\n  enum\r\n  {\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // generate rule name IDs
  cpID = cpIDNames;
  for(ul = 0; ul < ulRuleCount; ul++)
  {
    if(*cpID == 0){break;}
    ulBufLen = (ulong)sprintf(caBuf, "    %s=%lu,\r\n", cpID, ul);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    cpID += strlen(cpID) + 1;
  }
  ulBufLen = (ulong)sprintf(caBuf, "    ID_COUNT=%lu\r\n", ul);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  } RULENAMEIDS;\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // class function declarations
  ulBufLen = (ulong)sprintf(caBuf, "public:\r\n\r\n  // CLASS FUNCTIONS\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  %s();\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ~%s();\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulSyntaxAnalysis(ulong ulStartRule, uchar* ucpSrc, ulong ulSrcLen, void* vpData = NULL, ulong ulTrace = APG_FALSE);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulSemanticAnalysis(void* vpData);\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // call back function declarations
  ulBufLen = (ulong)sprintf(caBuf, "  // CALL BACK FUNCTIONS\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  cpID = cpTransNames;
  for(ul = 0; ul < ulRuleCount; ul++)
  {
    if(*cpID == 0){break;}
    ulBufLen = (ulong)sprintf(caBuf, "  static ulong %s(void* vpData, ulong ulState, ulong ulOffset, ulong ulLen);\r\n", cpID);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    cpID += strlen(cpID) + 1;
  }

  // generate the uniquely named pointer to the parser context
  ulBufLen = (ulong)sprintf(caBuf, "\r\nprivate:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  static void* m_vpParserInit;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  APG_BSTR m_sSrc;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\npublic:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  void* m_vpParserCtx;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  void* m_vpMemCtx;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  APG_PARSER_STATE m_sState;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  APG_PARSER_STATS m_sStats;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // class post markers
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s\r\n%s", cpClassPostBeg, cpClassPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // end of class declaration
  ulBufLen = (ulong)sprintf(caBuf, "\r\n};");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // user post-header markers
  ulBufLen = (ulong)sprintf(caBuf, "\r\n\r\n%s\r\n", cpHeaderPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpHeaderPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // complete the file
  ulBufLen = (ulong)sprintf(caBuf, "#endif // %s_h\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Creates a C++ language source file in a memory buffer (a byte vector)
//
// FUNCTION: static void vGenerateCppSource
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpCodeFileCtx
//              : pinter to a byte-element vector component context 
//
//      ARG: 3) APG_EXEHDR* spExe
//              : pointer to an executable data struct
//
//      ARG: 4) char* cpParserName
//              : the name of the parser
//
//      ARG: 5) char* cpFileName
//              : the name of the output file
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenerateCppSource(GENCTX* spData, void* vpCodeFileCtx, APG_EXEHDR* spExe, char* cpClassName, char* cpFileName)
{
  uchar*            ucpStrTbl;
  char*             cpRuleName;
  ulong             ulRuleCount = spData->ulRuleCount;
  APG_RULES*        spRule;
  ulong             ul;
  ulong             ulTransCount;
  char*             cpTrans;
  struct tm*        newTime;
  char*             cpAsciiTime;
  time_t            szClock;
  char              caBuf[APG_PAGE];
  ulong             ulBufLen;
  ulong             ulWriteUlongs;
  GRAMMAR_METRICS*  spGMetrics = &spData->sGrammarMetrics;
  
  // write the intro
  time( &szClock );
  newTime = localtime( &szClock );
  cpAsciiTime = asctime(newTime);
  ulBufLen = (ulong)sprintf(caBuf, "/********************************************************************\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // insert the GPL copyright notice
  ulBufLen = strlen(cpGPLNotice);
  vpVecPushn(vpCodeFileCtx, (void*)cpGPLNotice, ulBufLen);
  
  // NOTE: cpAsciiTime ends in \n only
  ulBufLen = (ulong)sprintf(caBuf, "	 created: ");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)strlen(cpAsciiTime);
  vpVecPushn(vpCodeFileCtx, (void*)cpAsciiTime, ulBufLen - 1);
  vpVecPushn(vpCodeFileCtx, (void*)"\r\n", 2);

  ulBufLen = (ulong)sprintf(caBuf, "	filename: %s\r\n", cpFileName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	  author: Lowell Thomas\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          lowell@coasttocoastresearch.com\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	          http://www.coasttocoastresearch.com\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "	 purpose: Implementation of parser class %s.\r\n\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*********************************************************************/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#include \"Apg.h\"\r\n\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpDefinesPreBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpDefinesPreEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "#include \"%s.h\"\r\n\r\n", &cpClassName[3]);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpDefinesPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n\r\n", cpDefinesPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // write general info
  ulBufLen = (ulong)sprintf(caBuf, "//  GENERAL INFORMATION:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  {
    ulong ulTotal = spExe->ulLength + spExe->ulRuleCount * sizeof(PFN_CALLBACK);
    double fKB = (double)ulTotal / 1024.0;
    double fMB = (double)ulTotal / (1024.0 * 1024.0);
    ulBufLen = (ulong)sprintf(caBuf, "//       static data = %lu (bytes)\r\n", ulTotal);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    ulBufLen = (ulong)sprintf(caBuf, "//                   = %.2E (KB)\r\n", fKB);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    ulBufLen = (ulong)sprintf(caBuf, "//                   = %.2E (MB)\r\n", fMB);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  ulBufLen = (ulong)sprintf(caBuf, "//   character count = %lu (grammar alphabet character set)\r\n", spData->ulCharSetSize);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "// unused characters = %lu (grammar alphabet character set)\r\n", spData->ulCharSetUnused);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//        rule count = %lu\r\n", spData->ulRuleCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//      opcode count = %lu\r\n", spData->ulOpcodeCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         ALT count = %lu(%lu) (alternations)\r\n", spGMetrics->ulALT, spGMetrics->ulAlternates);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         CAT count = %lu(%lu) (concatenations)\r\n", spGMetrics->ulCAT, spGMetrics->ulConcats);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         RNM count = %lu\r\n", spGMetrics->ulRNM);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         REP count = %lu\r\n", spGMetrics->ulREP);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         PRD count = %lu\r\n", spGMetrics->ulPRD);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TRG count = %lu\r\n", spGMetrics->ulTRG);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TBS count = %lu\r\n", spGMetrics->ulTBS);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         TLS count = %lu\r\n", spGMetrics->ulTLS);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//      string table = %lu (bytes)\r\n", spExe->sStrTbl.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//           opcodes = %lu (bytes)\r\n", spExe->sOpcodes.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "//         rule list = %lu (bytes)\r\n", spExe->sRuleList.ulLen);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  if(spData->ulPPT)
  {
    ulBufLen = (ulong)sprintf(caBuf, "//         PPT bytes = %lu\r\n", spExe->sPPT.ulLen);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  else
  {
    ulBufLen = (ulong)sprintf(caBuf, "//              PPTs = <not used>\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }

  // write the call back function declarations
//  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
//  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
//  ulBufLen = (ulong)sprintf(caBuf, "// CALL BACK FUNCTION DECLARATIONS\r\n");
//  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
  ulTransCount = 0;
  for(ul = 0; ul < ulRuleCount; ul++)
  {
    if(*cpTrans == 0){break;}
//    ulBufLen = (ulong)sprintf(caBuf, "ulong %s::%s(void* vpData, ulong ulState, ulong ulOffset, ulong ulLen);\r\n", cpClassName, cpTrans);
//    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    cpTrans += strlen(cpTrans) + 1;
    ++ulTransCount;
  }

  // write the call back function pointers
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "// CALL BACK FUNCTION POINTERS\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "static ulong g_ulCallbacks = %lu;\r\n", ulTransCount);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  if(ulTransCount > 0)
  {
    ulBufLen = (ulong)sprintf(caBuf, "static PFN_CALLBACK g_pfnCallbacks[%d] = \r\n{\r\n", ulTransCount);
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
    
    cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
    for(ul = 0; ul < ulRuleCount; ul++)
    {
      if(*cpTrans == 0){break;}
      ulBufLen = (ulong)sprintf(caBuf, "  &%s::%s,\r\n", cpClassName, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      cpTrans += strlen(cpTrans) + 1;
    }
    
    // finish call back function pointers
    ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  else
  {
    // handle case of zero call backs
    ulBufLen = (ulong)sprintf(caBuf, "static PFN_CALLBACK g_pfnCallbacks[1];\r\n");
    vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  }
  
  // executable data initialization
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// EXECUTABLE DATA\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulWriteUlongs = spExe->ulLength / sizeof(ulong);
  ulBufLen = (ulong)sprintf(caBuf, "static ulong g_ulaExeData[%d] = \r\n{\r\n", ulWriteUlongs);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // output the data
  vWriteCharacterTable(vpCodeFileCtx, (ulong*)spExe, ulWriteUlongs);
  
  // finish executable data initialization
  ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  // generate the unique parser named context pointer
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// UNIQUELY NAMED PARSER INITIALIZATION DATA\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "static APG_PARSER_INIT g_sParserInit = \r\n{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  ulBufLen = (ulong)sprintf(caBuf, "  0,\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  (void*)g_pfnCallbacks,\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  (void*)g_ulaExeData\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // finish unique parser named context pointer
  ulBufLen = (ulong)sprintf(caBuf, "};\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "void* %s::m_vpParserInit = (void*)&g_sParserInit;\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // the constructor
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s::%s()\r\n", cpClassName, cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulReturn = APG_FALSE;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulTest;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // get memory and error contexts\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_vpMemCtx = NULL;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_vpParserCtx = NULL;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_vpMemCtx = vpMemCtor();\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!m_vpMemCtx){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // initialize the parser\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulTest = ulParserInit(m_vpParserInit, m_vpMemCtx);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!ulTest){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_vpParserCtx = vpParserCtor(m_vpParserInit, m_vpMemCtx);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!m_vpParserCtx){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // success\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulReturn = APG_TRUE;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "APG_EXIT:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!ulReturn)\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  {\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // clean up resources\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    if(m_vpParserCtx){vParserDtor(m_vpParserCtx); m_vpParserCtx = NULL;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    if(m_vpMemCtx){vMemDtor(m_vpMemCtx); m_vpMemCtx = NULL;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  }\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // the destructor
  ulBufLen = (ulong)sprintf(caBuf, "%s::~%s()\r\n", cpClassName, cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // clean up resources\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(m_vpParserCtx){vParserDtor(m_vpParserCtx);}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(m_vpMemCtx){vMemDtor(m_vpMemCtx);}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // Syntax
  ulBufLen = (ulong)sprintf(caBuf, "\r\nulong %s::ulSyntaxAnalysis(ulong ulStartRule, uchar* ucpSrc, ulong ulSrcLen, void* vpData, ulong ulTrace)\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulReturn = APG_FALSE;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulong ulTest;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_sSrc.ucpStr = ucpSrc;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  m_sSrc.ulLen = ulSrcLen;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulTest = ulParserSyntaxAnalysis(m_vpParserCtx, 0, m_sSrc.ucpStr, m_sSrc.ulLen, vpData, ulTrace);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!ulTest){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulTest = ulParserState(m_vpParserCtx, &m_sState);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!ulTest){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulTest = ulParserStats(m_vpParserCtx, &m_sStats);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  if(!ulTest){goto APG_EXIT;}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  // success\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  ulReturn = APG_TRUE;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "APG_EXIT:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  return ulReturn;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // Semantics
  ulBufLen = (ulong)sprintf(caBuf, "\r\nulong %s::ulSemanticAnalysis(void* vpData)\r\n", cpClassName);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "{\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  return ulParserSemanticAnalysis(m_vpParserCtx, vpData);\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  

  // switch template for call back functions
  ulBufLen = (ulong)sprintf(caBuf, "\r\n// switch template\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "/*\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  switch(ulState)\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  {\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_PRE:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_NOMATCH:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_EMPTY:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SYN_MATCH:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SEM_PRE:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  case SEM_POST:\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "    break;\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "  }\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "*/\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  
  ulBufLen = (ulong)sprintf(caBuf, "\r\n%s\r\n", cpCallBackPostBeg);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
  ulBufLen = (ulong)sprintf(caBuf, "%s\r\n", cpCallBackPostEnd);
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

  // generate call back functions
  spRule = (APG_RULES*)vpVecFront(spData->vpRuleCtx);
  cpTrans = (char*)vpVecFront(spData->vpTransNamesCtx);
  ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  for(ul = 0; ul < ulRuleCount; ul++, spRule++)
  {
    if(spRule->usSyntax || spRule->usSemantics)
    {
      // add a call back function
      cpRuleName = (char*)ucpStrTbl + (ulong)spRule->cpRuleName;
      ulBufLen = (ulong)sprintf(caBuf, "\r\n// call back function for rule \"%s\"\r\n", cpRuleName);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "ulong %s::%s(void* vpData, ulong ulState, ulong ulOffset, ulong ulLen)\r\n", cpClassName, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf,  "{\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "  ulong ulReturn = PFN_OK;\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "%s.%s\r\n", cpCallBackBeg, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "%s.%s\r\n", cpCallBackEnd, cpTrans);
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "  return ulReturn;\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      ulBufLen = (ulong)sprintf(caBuf, "}\r\n");
      vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);
      cpTrans += strlen(cpTrans) + 1;
    }

  }
  ulBufLen = (ulong)sprintf(caBuf, "\r\n");
  vpVecPushn(vpCodeFileCtx, (void*)caBuf, ulBufLen);

}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Read user-written code from within the specially designated
//              comment boundaries and save it to a simple database.
//
// FUNCTION: static void vSaveUserCode
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpSavedUserCodeCtx
//              : byte-element vector to copy the user-written code to
//
//      ARG: 3) char* cpSrc
//              : pointer to the source code
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vSaveUserCode(GENCTX* spData, void* vpSavedUserCodeCtx, char* cpSrc)
{
  char* cpLineEnd = "\r\n";
  char* cpBeg = cpSrc;
  char* cpEnd;
  char* cpUserCode;
  char* cpKey;
  ulong ulKeyLen;
  char  caMatch[256];
  ulong ulCodeLen;
  void* vpBase;
  size_t iLen;
  DBENTRY* spEntry;

  while(cpBeg)
  {
    // "//{{" comment must begin at first character of a newline
    // (can't be first line of the file)
    cpBeg = strstr(cpBeg, "\r\n//{{");
    if(cpBeg)
    {
      // find the database key
      cpBeg += 2;
      cpEnd = strstr(cpBeg, cpLineEnd);
      GENASSERT(cpEnd);
      cpKey = cpBeg;
      cpUserCode = cpEnd + 2;
      ulKeyLen = (ulong)(cpUserCode - cpKey);
      GENASSERT((ulKeyLen < 256));
      
      // find the matching end-of-user-code
      memcpy((void*)caMatch, (void*)cpKey, ulKeyLen);
      caMatch[2] = '}';
      caMatch[3] = '}';
      caMatch[ulKeyLen] = 0;
      cpBeg = strstr(cpUserCode, caMatch);
      if(cpBeg)
      {
        ulCodeLen = (ulong)(cpBeg - cpUserCode);
        if(ulCodeLen)
        {
          // put a pointer to the user code on the list
          spEntry = (DBENTRY*)vpVecPush(vpSavedUserCodeCtx, NULL);
          spEntry->cpKey = cpKey;
          spEntry->ulKeyLen = ulKeyLen;
          spEntry->cpValue = cpUserCode;
          spEntry->ulValueLen = ulCodeLen;
        }
        
        // reset for next search
        cpBeg = cpUserCode + (ulCodeLen + ulKeyLen);
        cpBeg += 2;
      }
      // else error - matching end not found - skip this section
    }
  }

  // sort the list
  vpBase = vpVecFront(vpSavedUserCodeCtx);
  iLen = (size_t)ulVecSize(vpSavedUserCodeCtx);
  spEntry = (DBENTRY*)vpBase;
  qsort(vpBase, iLen, sizeof(DBENTRY), iCompare);

APG_EXIT:
  return;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Find the specially-designate comment boundaries and restore
//              any user-written code that may have been saved for this section.
//
// FUNCTION: static void vRestoreUserCode
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpRestoredFileCtx
//              : byte-element vector to copy the code + user-written code to
//
//      ARG: 3) void* vpCodeFileCtx
//              : byte-element vector to with the generated code
//
//      ARG: 4) void* vpSavedUserCodeCtx
//              : byte-element vector holding the previously-saved the user-written code
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vRestoreUserCode(GENCTX* spData, void* vpRestoredFileCtx, void* vpCodeFileCtx, void* vpSavedUserCodeCtx)
{
  char* cpSrc = (char*)vpVecFront(vpCodeFileCtx);
  ulong ulSrcLen = ulVecSize(vpCodeFileCtx);
  char* cpLineEnd = "\r\n";
  char* cpReplace = cpSrc;
  char* cpSrcEnd = cpSrc + ulSrcLen;
  ulong ulReplaceLen;
  char* cpBeg = cpSrc;
  char* cpEnd;
  char* cpUserCode;
  char* cpKey;
  ulong ulKeyLen;
  char  caMatch[256];
  void* vpBase = vpVecFront(vpSavedUserCodeCtx);
  size_t iLen = (size_t)ulVecSize(vpSavedUserCodeCtx);
  DBENTRY sKey;
  DBENTRY* spEnd;
  DBENTRY* spValue;
  ulong ulSaveCount;

  // mark all items in database as not found
  spValue = (DBENTRY*)vpBase;
  spEnd = spValue + iLen;
  for(; spValue < spEnd; ++spValue)
  {
    spValue->ulFound = APG_FALSE;
  }

  while(cpBeg)
  {
    // "//{{" comment must begin at first character of a newline
    // (can't be first line of the file)
    cpBeg = strstr(cpBeg, "\r\n//{{");
    if(cpBeg)
    {
      // find the database key
      cpBeg += 2;
      cpEnd = strstr(cpBeg, cpLineEnd);
      GENASSERT(cpEnd);
      cpKey = cpBeg;
      cpUserCode = cpEnd + 2;
      ulKeyLen = (ulong)(cpUserCode - cpKey);
      GENASSERT((ulKeyLen < 256));
      
      // find the matching end-of-user-code
      memcpy((void*)caMatch, (void*)cpKey, ulKeyLen);
      caMatch[2] = '}';
      caMatch[3] = '}';
      caMatch[ulKeyLen] = 0;
      cpBeg = strstr(cpUserCode, caMatch);
      if(cpBeg)
      {
        // look up the user code for this section
        sKey.cpKey = cpKey;
        sKey.ulKeyLen = ulKeyLen;
        spValue = bsearch((void*)&sKey, vpBase, iLen, sizeof(DBENTRY), iCompare);
        if(spValue)
        {
          // replace everything up to the user code
          ulReplaceLen = (ulong)(cpUserCode - cpReplace);
          vpVecPushn(vpRestoredFileCtx, (void*)cpReplace, ulReplaceLen);

          // replace the user code
          vpVecPushn(vpRestoredFileCtx, (void*)spValue->cpValue, spValue->ulValueLen);

          // mark value as found
          spValue->ulFound = APG_TRUE;
          
          // update the beginning of the next replacement section
          cpReplace = cpBeg;
        }
        
        // reset for next search
        cpBeg = strstr(cpBeg, cpLineEnd);
        cpBeg += 2;
      }
      // else error - matching end not found - skip this section
    }
  }

  // write the remainder of the file
  ulSrcLen = (ulong)(cpSrcEnd - cpReplace);
  vpVecPushn(vpRestoredFileCtx, (void*)cpReplace, ulSrcLen);

  // add any unrestored user code
  spValue = (DBENTRY*)vpBase;
  spEnd = spValue + iLen;
  ulSaveCount = 0;
  for(; spValue < spEnd; ++spValue)
  {
    if(spValue->ulFound == APG_FALSE)
    {
      if(ulSaveCount == 0)
      {
        // write a beginning unique user code comment
        vpVecPushn(vpRestoredFileCtx, (void*)"#if 0\r\n", 7);
      }

      // save the user code
      cpUserCode = (char*)vpVecPushn(vpRestoredFileCtx, (void*)spValue->cpKey, spValue->ulKeyLen);
      vpVecPushn(vpRestoredFileCtx, (void*)spValue->cpValue, spValue->ulValueLen);
      ++ulSaveCount;

      // modify the comment rule name
      cpUserCode[2] = 32;
      cpUserCode[3] = 32;
      cpUserCode[4] = 32;
      cpUserCode[5] = 32;
      cpUserCode[6] = 32;
      cpUserCode[7] = 32;
      cpUserCode[8] = 32;
      cpUserCode[9] = 32;
      cpUserCode[10] = 32;
      cpUserCode[11] = 32;
      cpUserCode[12] = 32;
    }
  }

  if(ulSaveCount)
  {
    // write the ending unique user code comment
    vpVecPushn(vpRestoredFileCtx, (void*)"#endif\r\n", 8);
  }

APG_EXIT:
  return;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: From the list of grammar rules, generate a list of
//              call back function names.
//
// FUNCTION: static void vGenCallBackFunctionNames
//
//      ARG: 1) void* vpVecCtx
//              : byte-vector to recieve a list of null-terminated
//                call back function names
//
//      ARG: 2) APG_RULES* spRules
//              : pointer to the rule list
//
//      ARG: 3) ulong ulRuleCount
//              : the number of rules in the list
//
//      ARG: 4) char* cpStrTbl
//              : pointer to the generator's string table
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenCallBackFunctionNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl)
{
  char* cpRuleName;
  char* cpTrans;
  APG_RULES* spRule;
  APG_RULES* spRuleEnd;

  // generate a list of translated (pfn_* translator function) names
  spRule = spRules;
  spRuleEnd = spRule + ulRuleCount;
  for(; spRule < spRuleEnd; ++spRule)
  {
    if(spRule->usSyntax || spRule->usSemantics)
    {
      cpRuleName = cpStrTbl + (ulong)spRule->cpRuleName;
      vpVecPushn(vpVecCtx, (void*)"pfn_", 4);
      cpTrans = (char*)vpVecPushn(vpVecCtx, (void*)cpRuleName, strlen(cpRuleName) + 1);
      while(*cpTrans != 0)
      {
        if(*cpTrans == '-')
        {
          *cpTrans = '_';
        }
        cpTrans++;
      }
    }
  }
  
  // end the list with a null-term
  cpTrans = (char*)vpVecPush(vpVecCtx, NULL);
  *cpTrans = 0;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: From the list of grammar rules, generate a list of
//              C language rule name identifiers.
//
// FUNCTION: static void vGenRuleIDNames
//
//      ARG: 1) void* vpVecCtx
//              : byte-vector to recieve a list of null-terminated
//                call back function names
//
//      ARG: 2) APG_RULES* spRules
//              : pointer to the rule list
//
//      ARG: 3) ulong ulRuleCount
//              : the number of rules in the list
//
//      ARG: 4) char* cpStrTbl
//              : pointer to the generator's string table
//
//      ARG: 5) char* cpParserName
//              : the parser name
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenRuleIDNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl, char* cpParserName)
{
  char* cpRuleName;
  char* cpID;
  APG_RULES* spRule;
  APG_RULES* spRuleEnd;
  ulong ulIDOffset;
  ulong ulParserNameLen = (ulong)strlen(cpParserName);
  
  // generate a list of translated rule name IDs
  spRule = spRules;
  spRuleEnd = spRule + ulRuleCount;
  for(; spRule < spRuleEnd; ++spRule)
  {
    ulIDOffset = ulVecSize(vpVecCtx);
    cpRuleName = cpStrTbl + (ulong)spRule->cpRuleName;
    vpVecPushn(vpVecCtx, (void*)"RULE_", 5);
    vpVecPushn(vpVecCtx, (void*)cpParserName, ulParserNameLen);
    vpVecPushn(vpVecCtx, (void*)"_", 1);
    vpVecPushn(vpVecCtx, (void*)cpRuleName, strlen(cpRuleName) + 1);
    cpID = (char*)vpVecFront(vpVecCtx) + ulIDOffset;
    while(*cpID != 0)
    {
      if(*cpID == '-')
      {
        *cpID = '_';
      }
      else
      {
        if(islower(*cpID))
        {
          *cpID = toupper(*cpID);
        }
      }
      cpID++;
    }
  }

  // end the names with a rule count
  ulIDOffset = ulVecSize(vpVecCtx);
  vpVecPushn(vpVecCtx, (void*)"RULE_COUNT_", 11);
  vpVecPushn(vpVecCtx, (void*)cpParserName, ulParserNameLen + 1);
  cpID = (char*)vpVecFront(vpVecCtx) + ulIDOffset;
  while(*cpID != 0)
  {
    if(*cpID == '-')
    {
      *cpID = '_';
    }
    else
    {
      if(islower(*cpID))
      {
        *cpID = toupper(*cpID);
      }
    }
    cpID++;
  }

  // end the list with a null-term
  cpID = (char*)vpVecPush(vpVecCtx, NULL);
  *cpID = 0;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: From the list of grammar rules, generate a list of
//              C++ language rule name identifiers.
//
// FUNCTION: static void vGenCppRuleIDNames
//
//      ARG: 1) void* vpVecCtx
//              : byte-vector to recieve a list of null-terminated
//                call back function names
//
//      ARG: 2) APG_RULES* spRules
//              : pointer to the rule list
//
//      ARG: 3) ulong ulRuleCount
//              : the number of rules in the list
//
//      ARG: 4) char* cpStrTbl
//              : pointer to the generator's string table
//
//      ARG: 5) char* cpParserName
//              : the parser name
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vGenCppRuleIDNames(void* vpVecCtx, APG_RULES* spRules, ulong ulRuleCount, char* cpStrTbl, char* cpParserName)
{
  char* cpRuleName;
  char* cpID;
  APG_RULES* spRule;
  APG_RULES* spRuleEnd;
  ulong ulIDOffset;
  ulong ulParserNameLen = (ulong)strlen(cpParserName);
  
  // generate a list of translated rule name IDs
  spRule = spRules;
  spRuleEnd = spRule + ulRuleCount;
  for(; spRule < spRuleEnd; ++spRule)
  {
    ulIDOffset = ulVecSize(vpVecCtx);
    cpRuleName = cpStrTbl + (ulong)spRule->cpRuleName;
    vpVecPushn(vpVecCtx, (void*)"ID_", 3);
    vpVecPushn(vpVecCtx, (void*)cpRuleName, strlen(cpRuleName) + 1);
    cpID = (char*)vpVecFront(vpVecCtx) + ulIDOffset;
    while(*cpID != 0)
    {
      if(*cpID == '-')
      {
        *cpID = '_';
      }
      else
      {
        if(islower(*cpID))
        {
          *cpID = toupper(*cpID);
        }
      }
      cpID++;
    }
  }
  
  // end the names with a rule count
  ulIDOffset = ulVecSize(vpVecCtx);
  vpVecPushn(vpVecCtx, (void*)"ID_MAX", 6);

  // end the list with a null-term
  cpID = (char*)vpVecPush(vpVecCtx, NULL);
  *cpID = 0;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Comparison function for a binary search of the simple
//              database of saved user-written code.
//
// FUNCTION: static int iCompare
//
//      ARG: 1) const void* vpLeft
//              : pointer to the left-hand side of the comparison
//
//      ARG: 2) const void* vpRight
//              : pointer to the right-hand side of the comparison
//
//   RETURN: -1 if left <= right
//            1 if left >= right
//            0 if left == right
//
//////////////////////////////////////////////////////////////////////////
static int iCompare(const void* vpLeft, const void* vpRight)
{
  const DBENTRY* spLeft = (DBENTRY*)vpLeft;
  const DBENTRY* spRight = (DBENTRY*)vpRight;
  
  if(spLeft->ulKeyLen < spRight->ulKeyLen) return -1;
  if(spLeft->ulKeyLen > spRight->ulKeyLen) return 1;
  return memcmp((void*)spLeft->cpKey, (void*)spRight->cpKey, spLeft->ulKeyLen);
}

// assumes that a printed item is never longer than a printed line
// i.e. MAXITEMLEN   < MAXLINELEN, for every item
#define MAXLINELEN 90
#define MAXITEMLEN 15
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Write an array of 32-bit integers as ASCII values that
//              can be used in a C/C++ language variable initialization statement.
//
// FUNCTION: static void vWriteCharacterTable
//
//      ARG: 1) void* vpCodeFileCtx
//              : pointer to the byte-element vector to recieve the ASCII format values
//
//      ARG: 2) ulong* ulpTable
//              : pointer to the array of integers to convert
//
//      ARG: 3) ulong ulTableSize
//              : number of integers in the array
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vWriteCharacterTable(void* vpCodeFileCtx, ulong* ulpTable, ulong ulTableSize)
{
  char caLine[MAXLINELEN + 5];
  char caItem[MAXITEMLEN + 5];
  ulong ulPrevLine;
  ulong ulThisItem;
  ulong* ulpBeg = ulpTable;
  ulong* ulpEnd = ulpBeg + ulTableSize;

  ulPrevLine = 2;
  caLine[0] = 32;
  caLine[1] = 32;
  caLine[2] = 0;
  for(; ulpBeg < ulpEnd; ++ulpBeg)
  {
    ulThisItem = sprintf(caItem, "0x%08X,", *ulpBeg);
    if(ulThisItem + ulPrevLine > MAXLINELEN)
    {
      caLine[ulPrevLine++] = '\r';
      caLine[ulPrevLine++] = '\n';
      vpVecPushn(vpCodeFileCtx, (void*)caLine, ulPrevLine);
      ulPrevLine = 2;
      caLine[0] = 32;
      caLine[1] = 32;
      caLine[2] = 0;
    }

    strcat(caLine, caItem);
    ulPrevLine += ulThisItem;
  }

  if(ulPrevLine)
  {
    caLine[ulPrevLine++] = '\r';
    caLine[ulPrevLine++] = '\n';
    vpVecPushn(vpCodeFileCtx, (void*)caLine, ulPrevLine);
  }
}
