/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:56
  filename: D:\Projects\Apgv5_0\Generator\Grammar.c
   purpose: Reads, parses and translates the input grammar file.
  
*********************************************************************/
#include "Generator.h"
#include "Lexer.h"
#include "Parser.h"

static ulong ulReducedOpcodes(GENCTX* spData);

#define CR          13
#define LF          10
#define SP          32
#define TAB         9
#define IS_ASCII(c)         (((uchar)(c) >= 32) && ((uchar)(c) < 128))

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: validate, translate and catalog lines of the grammar file
//              all lines should end with CRLF
//        NOTE: A common and easy error to make is to provide a grammar
//              file with no CRLF at the end of the last line.
//              This function will correct that.
//              Caller must insure that the grammar file buffer has at
//              least two extra characters to provide for this.
//
// FUNCTION: ulong ulGrammarValidate
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) APG_BSTR* spGrammar
//              : 
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulGrammarValidate(GENCTX* spData, APG_BSTR* spGrammar)
{
  ulong ulReturn = APG_FALSE;
  char* cpSrc;
  char* cpSrcBeg;
  char  cSave;
  ulong ulLen;
  ulong ulChar;
  ulong ulLine;
  ulong ulLineChar;
  ulong ulLineBeg;
  APG_BSTR* spBloc;
  DASSERT(spData && spData->vpValidate == (void*)spData);

  // validate the grammar file pointers
  if(!(spGrammar && spGrammar->ucpStr && spGrammar->ulLen > 1))
  {
    printf("Grammar File: error - grammar file is missing, empty or null pointer\n");
    goto APG_EXIT;
  }

  // check for CRLF at end of last line
  cpSrcBeg = (char*)spGrammar->ucpStr;
  DASSERT(cpSrcBeg[spGrammar->ulLen] == 0);
  cpSrc = &cpSrcBeg[spGrammar->ulLen - 2];
  if(!(cpSrc[0] == CR && cpSrc[1] == LF))
  {
    // add a CRLF to end of file
    cpSrc[2] = CR;
    cpSrc[3] = LF;
    cpSrc[4] = 0;
    if(spData->sConfig.ulDisplayWarnings)
    {
      printf("Grammar File: warning - CRLF added to end of grammar, file extended from %lu to %lu characters\n",
        spGrammar->ulLen, spGrammar->ulLen + 2);
    }
    spGrammar->ulLen += 2;
  }

  // scan file - validate, convert HTAB to SPACE, catalog lines
  ulReturn = APG_TRUE;
  ulLen = spGrammar->ulLen;
  cpSrc = cpSrcBeg;
  ulChar = 0;
  ulLine = 1;
  ulLineChar = 0;
  ulLineBeg = 0;
  while(ulChar < ulLen)
  {
    // convert tabs
    if(cpSrc[ulChar] == TAB){cpSrc[ulChar] = SP;}

    // check for line end
    if(cpSrc[ulChar] == CR)
    {
      if(cpSrc[ulChar + 1] == LF)
      {
        // good line end
        // catalog the line
        spBloc = (APG_BSTR*)vpVecPush(spData->vpLinesCtx, NULL);
        spBloc->ucpStr = (uchar*)&cpSrcBeg[ulChar - ulLineChar];
        spBloc->ulLen = ulLineChar;

        // continue with next line
        ulChar += 2;       // skip CRLF
        ++ulLine;
        ulLineChar = 0; // reset for next line
        ulLineBeg = ulChar;
      }
      else
      {
        // report error
        ulReturn = APG_FALSE;
        cSave = cpSrcBeg[ulChar];
        cpSrcBeg[ulChar] = 0;
        printf("\n%s\n", &cpSrcBeg[ulLineBeg]);
        cpSrcBeg[ulChar] = cSave;
        printf("Grammar File: error - CR found without LF at line, char[%lu, %lu]\n",
          ulLine, ulChar);
        
        // continue with next line
        ++ulLine;
        ++ulChar;
        ulLineChar = 0; // reset for next line
        ulLineBeg = ulChar;
      }

      continue;
    }

    // check for ASCII
    if(!IS_ASCII(cpSrc[ulChar]))
    {
      // report error
      ulReturn = APG_FALSE;
      if(cpSrc[ulChar] == LF)
      {
        cSave = cpSrcBeg[ulChar];
        cpSrcBeg[ulChar] = 0;
        printf("\n%s\n", &cpSrcBeg[ulLineBeg]);
        cpSrcBeg[ulChar] = cSave;
        printf("Grammar File: error - LF found without CR at line, char[%lu, %lu]\n",
          ulLine, ulChar);
        
        // continue with next line
        ++ulLine;
        ++ulChar;
        ulLineChar = 0; // reset for next line
        ulLineBeg = ulChar;
        continue;
      }
      else
      {
        cSave = cpSrcBeg[ulChar];
        cpSrcBeg[ulChar] = 0;
        printf("\n%s\n", &cpSrcBeg[ulLineBeg]);
        cpSrcBeg[ulChar] = cSave;
        printf("Grammar File: error - non-ASCII character[%u] found at line, char[%lu, %lu]\n",
          (uchar)cpSrc[ulChar], ulLine, ulChar);
      }
    }

    // increment to next character
    ++ulChar;
    ++ulLineChar;
  }
 
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: looks for a rule name in the rule list
//              does a brute force, case-insensitive linear search
//
// FUNCTION: ulong ulGrammarRuleLookup
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) const uchar* ucpSrc
//              : pointer to the grammar file source
//
//      ARG: 3) ulong ulSrcLen
//              : number of bytes in the grammar file source
//
//   RETURN: the rule number on success
//           UNDEFINED if the rule name is not found
//
//////////////////////////////////////////////////////////////////////////
ulong ulGrammarRuleLookup(GENCTX* spData, const uchar* ucpSrc, ulong ulSrcLen)
{
  ulong ulReturn = UNDEFINED;
  uchar* ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  INT_RULE_LIST* spRuleList = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
  INT_RULE_LIST* spBeg = spRuleList;
  INT_RULE_LIST* spEnd = spBeg + spData->ulRuleCount;
  ulong ulIndex = 0;
  for(; spBeg < spEnd; ++spBeg, ++ulIndex)
  {
    // compare name lengths
    if(ulSrcLen == spBeg->ulNameLen)
    {
      // compare lower case characters
      const uchar* ucpR = ucpSrc;
      uchar* ucpL = ucpStrTbl + spBeg->ulNameOffset;
      uchar* ucpEnd = ucpL + ulSrcLen;
      for(; ucpL < ucpEnd; ++ucpL, ++ucpR)
      {
        uchar ucL = *ucpL;
        uchar ucR = *ucpR;
        if(ucL >= 65 && ucL <= 90){ucL += 32;}
        if(ucR >= 65 && ucR <= 90){ucR += 32;}
        if(ucL != ucR){break;}
      }
      
      if(ucpL == ucpEnd)
      {
        // match found
        ulReturn = ulIndex;
        break;
      }
    }
    // else - no match
  }
  
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
//  INTERPERT DIRECTIVES IN THE GRAMMAR FILE
//////////////////////////////////////////////////////////////////////////
static char* SECTION_GRAMMAR = "[SECTION GRAMMAR]";
static char* SECTION_SYNTAX = "[SECTION SYNTAX]";
static char* SECTION_SEMANTICS = "[SECTION SEMANTICS]";
static char* SECTION_INCLUDE = "[INCLUDE]";
static char* SECTION_EXCLUDE = "[EXCLUDE]";
static char* ALL_RULES = "ALL_RULES";
static char* LINE_END = "\r\n";
static ulong LINE_END_LEN = 2;
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Find a line in the grammar file with a section name
//              "[SECTION GRAMMAR]", etc.
//
// FUNCTION: void vFindSection
//
//      ARG: 1) char* cpSectionName
//              : section name to find
//
//      ARG: 2) APG_BSTR* spFile
//              : binary string defining the grammar file
//
//      ARG: 3) APG_BSTR* spSection
//              : output on success, binary string pointer to the section contents
//                NULL string (not a null pointer) otherwise
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vFindSection(char* cpSectionName, APG_BSTR* spFile, APG_BSTR* spSection)
{
  char* cpBeg;
  char* cpEnd;

  spSection->ucpStr = NULL;
  spSection->ulLen = 0;

  while(APG_TRUE)
  {
    // look for the section
    cpBeg  = strstr((char*)spFile->ucpStr, cpSectionName);
    if(!cpBeg){break;}

    // find the end of the section
    cpBeg = strstr(cpBeg, LINE_END);
    cpBeg += LINE_END_LEN;
    cpEnd = strstr(cpBeg, "[SECTION");
    if(!cpEnd)
    {
      // found an INCLUDE section, locate the end
      cpEnd = (char*)spFile->ucpStr + spFile->ulLen;
    }

    spSection->ucpStr = (uchar*)cpBeg;
    spSection->ulLen = (ulong)(cpEnd - cpBeg);
    break;
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Find the contents of a section's sub-section
//              [INCLUDE] or [EXCLUDE]
//
// FUNCTION: void vFindSubSection
//
//      ARG: 1) char* cpSectionName
//              :  sub-section name
//
//      ARG: 2) APG_BSTR* spSection
//              : binary string defining the section
//
//      ARG: 3) APG_BSTR* spSub
//              : on success, binary string defining the contents of the sub-section
//                NULL string (not a null pointer) on failure
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
void vFindSubSection(char* cpSectionName, APG_BSTR* spSection, APG_BSTR* spSub)
{
  char* cpBeg;
  char* cpEnd;
  
  spSub->ucpStr = NULL;
  spSub->ulLen = 0;
  
  while(APG_TRUE)
  {
    // look for the sub section name
    cpBeg  = strstr((char*)spSection->ucpStr, cpSectionName);
    if(!cpBeg){break;}

    // find the end of the section
    cpBeg = strstr(cpBeg, LINE_END);
    cpBeg += LINE_END_LEN;
    cpEnd = strstr(cpBeg, "[");
    if(!cpEnd)
    {
      // found an INCLUDE section, locate the end
      cpEnd = (char*)spSection->ucpStr + spSection->ulLen;
    }
    
    spSub->ucpStr = (uchar*)cpBeg;
    spSub->ulLen = (ulong)(cpEnd - cpBeg);
    break;
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Build a list of rule name indicators.
//              rule list indicators are 16-bit values that are
//              1 if the rule name is included in the list
//              0 if the rule name is not included in the list
//
// FUNCTION: void vBuildList
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) INT_RULE_LIST* spRuleList
//              : pointer to the list of intermediate rule names
//
//      ARG: 3) ulong ulRuleCount
//              : number of rule names in the list
//
//      ARG: 4) APG_BSTR* spInclude
//              : binary string defining the contents of an INCLUDE list
//
//      ARG: 5) APG_BSTR* spExclude
//              : binary string defining the contents of an EXCLUDE list
//
//      ARG: 6) uchar* ucpSynList
//              : pointer to the list of indicators
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vBuildList(GENCTX* spData, INT_RULE_LIST* spRuleList, ulong ulRuleCount, APG_BSTR* spInclude, APG_BSTR* spExclude, uchar* ucpSynList)
{
  char* cpBeg;
  char* cpEnd;
  char  cSave;
  char* cpAll;
  char* cpLine;
  char* cpCRLF;
  ulong ulIndex;

  // build the include list
  cpBeg = (char*)spInclude->ucpStr;
  if(cpBeg)
  {
    cpEnd = cpBeg + spInclude->ulLen;
    cSave = *cpEnd;
    *cpEnd = 0;
    cpAll = strstr(cpBeg, ALL_RULES);
    if(cpAll)
    {
      // add all 
      memset((void*)ucpSynList, 1, ulRuleCount);
    }
    else
    {
      // search through rule names
      memset((void*)ucpSynList, 0, ulRuleCount);
      cpLine = cpBeg;
      cpCRLF = strstr(cpLine, LINE_END);
      while(cpCRLF)
      {
        ulIndex = ulGrammarRuleLookup(spData, (uchar*)cpLine, (ulong)(cpCRLF - cpLine));
        if(ulIndex != UNDEFINED)
        {
          ucpSynList[ulIndex] = 1;
        }
        cpLine = cpCRLF + LINE_END_LEN;
        cpCRLF = strstr(cpLine, "\r\n");
      }
    }
    *cpEnd = cSave;
  }

  // build the exclude list
  cpBeg = (char*)spExclude->ucpStr;
  if(cpBeg)
  {
    cpEnd = cpBeg + spExclude->ulLen;
    cSave = *cpEnd;
    *cpEnd = 0;
    cpAll = strstr(cpBeg, ALL_RULES);
    if(cpAll)
    {
      // subtract all 
      memset((void*)ucpSynList, 0, ulRuleCount);
    }
    else
    {
      // search through rule names
      cpLine = cpBeg;
      cpCRLF = strstr(cpLine, LINE_END);
      while(cpCRLF)
      {
        ulIndex = ulGrammarRuleLookup(spData, (uchar*)cpLine, (ulong)(cpCRLF - cpLine));
        if(ulIndex != UNDEFINED)
        {
          ucpSynList[ulIndex] = 0;
        }
        cpLine = cpCRLF + LINE_END_LEN;
        cpCRLF = strstr(cpLine, LINE_END);
      }
    }
    *cpEnd = cSave;
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Find all of the grammar file sections
//
// FUNCTION: void vGrammarFileSections
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) APG_BSTR* spFile
//              : input, the entire grammar file
//
//      ARG: 3) APG_BSTR* spGrammar
//              : output, the contents of the grammar section
//
//      ARG: 4) APG_BSTR* spSyntax
//              : output, the contents of the syntax section
//
//      ARG: 5) APG_BSTR* spSemantics
//              : output, the contents of the semantics section
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vGrammarFileSections(GENCTX* spData, APG_BSTR* spFile, APG_BSTR* spGrammar, APG_BSTR* spSyntax, APG_BSTR* spSemantics)
{
  GENASSERT(spFile);
  GENASSERT(spFile->ucpStr);
  GENASSERT(spFile->ulLen);
  GENASSERT(spGrammar);

  vFindSection(SECTION_GRAMMAR, spFile, spGrammar);
  if(spGrammar->ucpStr == NULL || spGrammar->ulLen == 0)
  {
    // default grammar to entire file
    *spGrammar = *spFile;
    GENASSERT(spGrammar->ucpStr[spGrammar->ulLen] == 0);
  }
  else
  {
    // get the other sections, if any
    if(spSyntax)
    {
      vFindSection(SECTION_SYNTAX, spFile, spSyntax);
    }
    
    if(spSemantics)
    {
      vFindSection(SECTION_SEMANTICS, spFile, spSemantics);
    }

    // null-term syntax and semantic sections
    spGrammar->ucpStr[spGrammar->ulLen] = 0;
    if(spSyntax && spSyntax->ucpStr && spSyntax->ulLen)
    {
      spSyntax->ucpStr[spSyntax->ulLen] = 0;
    }
    
    if(spSemantics && spSemantics->ucpStr && spSemantics->ulLen)
    {
      spSemantics->ucpStr[spSemantics->ulLen] = 0;
    }
  }

APG_EXIT:
  return;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Adds syntax and semantics control indicators to the
//              intermediate rule list.
//
// FUNCTION: void vGrammarRuleNameControl
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) APG_BSTR* spSyntax
//              : contents of the SYNTAX section
//
//      ARG: 3) APG_BSTR* spSemantics
//              : contents of the SEMANTICS section
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vGrammarRuleNameControl(GENCTX* spData, APG_BSTR* spSyntax, APG_BSTR* spSemantics)
{
  ulong ulMemChk = ulMemCheckPoint(spData->vpMemCtx);
  INT_RULE_LIST* spRuleList = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
  INT_RULE_LIST* spRule;
  INT_RULE_LIST* spRuleEnd;
  ulong ulRuleCount = spData->ulRuleCount;
  uchar* ucpSynList;
  uchar* ucpSemList;
  uchar* ucpSyn;
  uchar* ucpSem;
  APG_BSTR sInclude;
  APG_BSTR sExclude;

  // create a single exit point
  while(APG_TRUE)
  {
    // allocate and initialize the lists to include all
    ucpSynList = (uchar*)vpMemAlloc(spData->vpMemCtx, ulRuleCount * sizeof(uchar));
    GENASSERT(ucpSynList);
    memset((void*)ucpSynList, (uchar)APG_TRUE, ulRuleCount * sizeof(uchar));
    ucpSemList = (uchar*)vpMemAlloc(spData->vpMemCtx, ulRuleCount * sizeof(uchar));
    GENASSERT(ucpSemList);
    memset((void*)ucpSemList, (uchar)APG_TRUE, ulRuleCount * sizeof(uchar));

    // get the syntax analysis call back list
    if(spSyntax && spSyntax->ucpStr && spSyntax->ulLen)
    {
      vFindSubSection(SECTION_INCLUDE, spSyntax, &sInclude);
      vFindSubSection(SECTION_EXCLUDE, spSyntax, &sExclude);
      vBuildList(spData, spRuleList, ulRuleCount, &sInclude, &sExclude, ucpSynList);
    }

    // get the semantic analysis call back list
    if(spSemantics && spSemantics->ucpStr && spSemantics->ulLen)
    {
      vFindSubSection(SECTION_INCLUDE, spSemantics, &sInclude);
      vFindSubSection(SECTION_EXCLUDE, spSemantics, &sExclude);
      vBuildList(spData, spRuleList, ulRuleCount, &sInclude, &sExclude, ucpSemList);
    }

    // success
    break;
  }

  // generate the rule list
  spRule = spRuleList;
  spRuleEnd = spRule + ulRuleCount;
  ucpSyn = ucpSynList;
  ucpSem = ucpSemList;
  for(; spRule < spRuleEnd; ++spRule, ++ucpSyn, ++ucpSem)
  {
    spRule->usSyntax = *ucpSyn;
    spRule->usSemantics = *ucpSem;
  }
  
APG_EXIT:
  // clean up memory allocations
  vMemFreeToCheckPoint(spData->vpMemCtx, ulMemChk);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Processes the intermediate rule list and extracts
//              the grammar metrics.
//
// FUNCTION: ulong ulGrammarMetrics
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: true
//
//////////////////////////////////////////////////////////////////////////
ulong   ulGrammarMetrics(GENCTX* spData)
{
  INT_RULE_LIST* spRule;
  INT_RULE_LIST* spRuleList = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
  INT_RULE_LIST* spEnd = spRuleList + spData->ulRuleCount;
  GRAMMAR_METRICS* spMetrics = &spData->sGrammarMetrics;
  OP_INT_ALL* spOp = (OP_INT_ALL*)vpVecFront(spData->vpIntOpcodesCtx);
  
  // get the character set metrics
  vGrammarCharacterSet(spData);
  spMetrics->ulCharSetSize = spData->ulCharSetSize;
  spMetrics->ulCharSetUsed = spData->ulCharSetUsed;
  spMetrics->ulRuleCount = spData->ulRuleCount;
  
  // count production opcodes (skip ALT(1), CAT(1) & REP(1,1))
  for(spRule = spRuleList; spRule < spEnd; ++spRule)
  {
    OP_INT_ALL* spOpEnd = spOp + spRule->ulOpcodeCount;
    for(; spOp < spOpEnd; ++spOp)
    {
      // skip unneeded opcodes
      if(spOp->ulOpId == OP_INT_ID_ALT && spOp->ulCount == 1){continue;}
      if(spOp->ulOpId == OP_INT_ID_CAT && spOp->ulCount == 1){continue;}
      if(spOp->ulOpId == OP_INT_ID_REP && spOp->ulMin == 1 && spOp->ulMax == 1){continue;}
      
      switch(spOp->ulOpId)
      {
      case OP_INT_ID_ALT:
        ++spMetrics->ulALT;
        //++spMetrics->ulNonTerminals;
        spMetrics->ulAlternates += spOp->ulCount;
        break;
      case OP_INT_ID_CAT:
        ++spMetrics->ulCAT;
        //++spMetrics->ulNonTerminals;
        spMetrics->ulConcats += spOp->ulCount;
        break;
      case OP_INT_ID_REP:
        ++spMetrics->ulREP;
        //++spMetrics->ulNonTerminals;
        break;
      case OP_INT_ID_RNM:
        ++spMetrics->ulRNM;
        //++spMetrics->ulNonTerminals;
        break;
      case OP_INT_ID_PRD:
        ++spMetrics->ulPRD;
        //++spMetrics->ulNonTerminals;
        break;
      case OP_INT_ID_TRG:
        ++spMetrics->ulTRG;
        //++spMetrics->ulTerminals;
        break;
      case OP_INT_ID_TBS:
        ++spMetrics->ulTBS;
        //++spMetrics->ulTerminals;
        break;
      case OP_INT_ID_TLS:
        ++spMetrics->ulTLS;
        //++spMetrics->ulTerminals;
        break;
      default:
        break;
      }
    }
  }

  if(spData->sConfig.ulDisplayMetrics)
  {
    vPrintGrammarMetrics(spData);
  }

  if(spData->sConfig.ulDisplayCharMap)
  {
    vPrintCharMap(spData);
  }
 
  return APG_TRUE;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Reads the grammar file into a memory buffer,
//              finds the grammar file sections,
//              validates the grammar format
//              initializes for grammar parsing
//
// FUNCTION: ulong ulGrammar
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong   ulGrammar(GENCTX* spData)
{
  ulong ulReturn = APG_FALSE;
  ulong ulTest;
  char* cpFileName = &spData->sConfig.caInput[0];

  // get a buffer and read the input grammar
  // NOTE: input buffer must be at least 2 greater than file size
  //       to allow for a possible line end correction
  // NOTE: a null terminator is added for the file section algorithm
  spData->sSource.ulLen = ulGetFileSize(cpFileName);
  GENASSERT(spData->sSource.ulLen);
  spData->sSource.ucpStr = (char*)vpMemAlloc(spData->vpMemCtx, spData->sSource.ulLen  + 10);
  GENASSERT(spData->sSource.ucpStr);
  spData->sSource.ulLen = ulGetFile(cpFileName, (void*)spData->sSource.ucpStr);
  GENASSERT(spData->sSource.ulLen);
  spData->sSource.ucpStr[spData->sSource.ulLen] = 0;
  
  // pull the grammar from the grammar section, if any
  vGrammarFileSections(spData, &spData->sSource, &spData->sGrammar, &spData->sSyntaxControl, &spData->sSemanticsControl);

  // validate the grammar file and replace htabs with spaces
  ulTest = ulGrammarValidate(spData, &spData->sGrammar);
  GENASSERT(ulTest);

  // initialize the erratum component for grammar error reporting
  spData->vpErrorsCtx = vpGrammarErrorsCtor(spData->vpMemCtx, spData->vpLinesCtx);
  GENASSERT(spData->vpErrorsCtx);
  spData->vpWarningsCtx = vpGrammarErrorsCtor(spData->vpMemCtx, spData->vpLinesCtx);
  GENASSERT(spData->vpWarningsCtx);

  // redundant for backward compatibility
  spData->ucpSrc = spData->sGrammar.ucpStr;
  spData->ulSrcLen = spData->sGrammar.ulLen;
  
  // print the source file
  if(spData->sConfig.ulDisplayGrammar)
  {
    printf("input grammar file: \"%s\"\n", cpFileName);
    vDisplaySourceLines(spData->sGrammar.ucpStr, stdout);
  }
  
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Grammar character set processing helper
//
// FUNCTION: static ulong* ulpUpdateMapX
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) ulong ulMax
//              : 
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static ulong* ulpUpdateMapX(GENCTX* spData, ulong ulMax)
{
  ulong ulCharSet = ulVecSize(spData->vpMapXCtx);
  if(ulMax >= ulCharSet)
  {
    // add the necessary bytes to the char set flags vector
    ulong ulAdd = ++ulMax - ulCharSet;
    ulong* ulpAdd = (ulong*)vpVecPushn(spData->vpMapXCtx, NULL, ulAdd);
    memset((void*)ulpAdd, 0xFF, ulAdd * sizeof(ulong));
  }
  
  // set the character set flags for this character range
  return (ulong*)vpVecFront(spData->vpMapXCtx);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Grammar character set processing helper
//
// FUNCTION: static void vSetTRGCharSetFlags
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) ulong ulMin
//              : 
//
//      ARG: 3) ulong ulMax
//              : 
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static void vSetTRGCharSetFlags(GENCTX* spData, ulong ulMin, ulong ulMax)
{
  ulong* ulpFlags = ulpUpdateMapX(spData, ulMax);
  ulong ulChar;
  for(ulChar = ulMin; ulChar <= ulMax; ++ulChar)
  {
    ulpFlags[ulChar] = ulChar;
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Grammar character set processing helper
//
// FUNCTION: static void vSetTBSCharSetFlags
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) uchar* ucpString
//              : 
//
//      ARG: 3) ulong ulLen
//              : 
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static void vSetTBSCharSetFlags(GENCTX* spData, uchar* ucpString, ulong ulLen)
{
  ulong* ulpFlags;
  uchar* ucpNext = ucpString;
  uchar* ucpEnd = ucpNext + ulLen;
  
  // scan all characters in this string
  for(; ucpNext < ucpEnd; ++ucpNext)
  {
    ulong ulChar = (ulong)*ucpNext;
    ulpFlags = ulpUpdateMapX(spData, ulChar);
    ulpFlags[ulChar] = ulChar;
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Grammar character set processing helper
//
// FUNCTION: static void vSetTLSCharSetFlags
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) uchar* ucpString
//              : 
//
//      ARG: 3) ulong ulLen
//              : 
//
//   RETURN: 
//
//////////////////////////////////////////////////////////////////////////
static void vSetTLSCharSetFlags(GENCTX* spData, uchar* ucpString, ulong ulLen)
{
  ulong* ulpFlags;
  uchar* ucpNext = ucpString;
  uchar* ucpEnd = ucpNext + ulLen;
  
  // scan all characters in this string
  for(; ucpNext < ucpEnd; ++ucpNext)
  {
    ulong ulChar = (ulong)*ucpNext;
    ulong ulMaxChar;
    if(ulChar >= 65 && ulChar <= 90)
    {
      // set both upper and lower case bits
      ulMaxChar = ulChar + 32;
      ulpFlags = ulpUpdateMapX(spData, ulMaxChar);
      ulpFlags[ulMaxChar] = ulMaxChar;
      ulpFlags[ulChar] = ulChar;
    }
    if(ulChar >= 97 && ulChar <= 122)
    {
      // set both upper and lower case bits
      ulMaxChar = ulChar;
      ulChar = ulMaxChar - 32;
      ulpFlags = ulpUpdateMapX(spData, ulMaxChar);
      ulpFlags[ulMaxChar] = ulMaxChar;
      ulpFlags[ulChar] = ulChar;
    }
    else
    {
      // just set the character
      ulpFlags = ulpUpdateMapX(spData, ulChar);
      ulpFlags[ulChar] = ulChar;
    }
  }
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: scan the terminal intermediate opcodes and map the character sets
//
// FUNCTION: void vGrammarCharacterSet
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
void vGrammarCharacterSet(GENCTX* spData)
{
  OP_INT_ALL* spIntOp;
  OP_INT_ALL* spIntOpEnd;
  ulong ulOpIndex;
  uchar* ucpStrTbl;
  ulong ulIndexX, ulIndexI;
  ulong* ulpMapX;

  // create the internal (I) and external (X) character maps
  spData->vpMapXCtx = vpVecCtor(spData->vpMemCtx, sizeof(ulong), 128);
  GENASSERT(spData->vpMapXCtx);
  spData->vpMapICtx = vpVecCtor(spData->vpMemCtx, sizeof(ulong), 128);
  GENASSERT(spData->vpMapICtx);

  // interate over all opcodes
  ucpStrTbl = (uchar*)vpVecFront(spData->vpStrTblCtx);
  spIntOp = (OP_INT_ALL*)vpVecFront(spData->vpIntOpcodesCtx);
  spIntOpEnd = spIntOp + ulVecSize(spData->vpIntOpcodesCtx);
  ulOpIndex = 0;
  for(; spIntOp < spIntOpEnd; ++spIntOp, ++ulOpIndex)
  {
    switch(spIntOp->ulOpId)
    {
      // just do statistics for non-terminal operators
    case OP_INT_ID_ALT:
    case OP_INT_ID_CAT:
    case OP_INT_ID_RNM:
    case OP_INT_ID_PRD:
    case OP_INT_ID_REP:
      break;
    case OP_INT_ID_TRG:
      {
        // update the character set flags
        vSetTRGCharSetFlags(spData, spIntOp->ulMin, spIntOp->ulMax);
      }
      break;
    case OP_INT_ID_TBS:
      {
        uchar* ucpString = ucpStrTbl + spIntOp->ulOffset;
        vSetTBSCharSetFlags(spData, ucpString, spIntOp->ulLength);
      }
      break;
    case OP_INT_ID_TLS:
      {
        uchar* ucpString = ucpStrTbl + spIntOp->ulOffset;
        vSetTLSCharSetFlags(spData, ucpString, spIntOp->ulLength);
      }
      break;
    default:
      GENASSERT(APG_FALSE);
      break;
    } // switch(spIntOp->alt.ulOpId)
  }   // for(; spIntOp < spIntOpEnd; ++spIntOp, ++ulOpIndex)

  // get the character set size
  spData->ulCharSetSize = ulVecSize(spData->vpMapXCtx);
  spData->ulEOFChar = spData->ulCharSetSize;
  spData->ulMaxChar = spData->ulCharSetSize - 1;
  spData->ulEOFSetSize = spData->ulCharSetSize + 1;
  spData->ulPPTSize1 = ulCeiling8(spData->ulEOFSetSize + 2);
  spData->ulPPTSize2 = ulCeiling4(spData->ulEOFSetSize + 1);

  // add an EOF character to the external map
  ulpMapX = (ulong*)vpVecPush(spData->vpMapXCtx, NULL);
  *ulpMapX = spData->ulCharSetSize;
  DASSERT(spData->ulEOFSetSize == ulVecSize(spData->vpMapXCtx));
    
  
  // generate the internal and external maps
  ulpMapX = (ulong*)vpVecFront(spData->vpMapXCtx);
  ulIndexI = 0;
  for(ulIndexX = 0; ulIndexX < spData->ulEOFSetSize; ++ulIndexX)
  {
    if(ulpMapX[ulIndexX] != UNDEFINED)
    {
      vpVecPush(spData->vpMapICtx, (void*)&ulpMapX[ulIndexX]);
      ulpMapX[ulIndexX] = ulIndexI;
      ++ulIndexI;
    }
  }

  spData->ulCharSetUsed = ulIndexI - 1;
  spData->ulCharSetUnused = spData->ulCharSetSize - spData->ulCharSetUsed;

  //NOTE: 12/08/06
  // possible future feature:
  // 1) transform terminal opcodes from external to internal character set
  // 2) transform input string from external to internal character set
  // 3) parse transformed input string
  // 4) transform AST back to external character set for semantic analysis
  // If a) internal set is much smaller than external set and 
  //    b) for some reason can't be reduced
  //    this may significantly reduce over all parsing performance.
  // I'm just guessing that this may be the case when parsing non-English languages
  // and the input string is in UNICODE.
//   vVecDtor(spData->vpMapXCtx);
//   spData->vpMapXCtx = NULL;

APG_EXIT:
  vVecDtor(spData->vpMapICtx);
  spData->vpMapICtx = NULL;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: parse the SABNF grammar into intermediate opcodes
//
// FUNCTION: ulong ulGrammarParser
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//      ARG: 2) void* vpLexer
//              : lexer parser component context
//
//      ARG: 3) void* vpParser
//              : token parser component context
//
//   RETURN: true on success
//           on success, generates a list of intermediate opcodes
//
//////////////////////////////////////////////////////////////////////////
ulong   ulGrammarParser(GENCTX* spData, void* vpLexer, void* vpParser)
{
  ulong             ulDisplayOpcodes = spData->sConfig.ulDisplayOpcodes;
  ulong             ulReturn = APG_FALSE;
  void*             vpParserCtx = NULL;
  uchar*            ucpTokIdx;
  ulong             ulTokIdxSize;
  APG_BSTR          sSrc;
  APG_PARSER_STATE  sState; 
  ulong             ulTest;
  
  // create the token table & index
  spData->vpTokTblCtx = vpVecCtor(spData->vpMemCtx, sizeof(SABNFTOKEN), 1024);
  GENASSERT(spData->vpTokTblCtx);
  spData->vpTokIdxCtx = vpVecCtor(spData->vpMemCtx, sizeof(uchar), 1024);
  GENASSERT(spData->vpTokIdxCtx);
  
  // create the string table
  spData->vpStrTblCtx = vpVecCtor(spData->vpMemCtx, sizeof(uchar), 1024);
  GENASSERT(spData->vpStrTblCtx);
  
  // create the rule list
  spData->vpIntRuleCtx = vpVecCtor(spData->vpMemCtx, sizeof(INT_RULE_LIST), 128);
  GENASSERT(spData->vpIntRuleCtx);
  
  //////////////////////////////////////////////////////////////////////////
  // lexical analysis of the grammar and translation to tokens
  //////////////////////////////////////////////////////////////////////////
  spData->ulLineNo = 1;
  ulTest = ulParserInit(vpLexer, spData->vpMemCtx);
  GENASSERT(ulTest);
  vpParserCtx = vpParserCtor(vpLexer, spData->vpMemCtx);
  GENASSERT(vpParserCtx);
  
  ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, spData->sGrammar.ucpStr, spData->sGrammar.ulLen, NULL, APG_FALSE);
  GENASSERT(ulTest);
  ulParserState(vpParserCtx, &sState);
  if(spData->sConfig.ulDisplayState)
  {
    vPrintState(&sState, "*** Grammar Lexer");
  }
  GENASSERT(sState.ulState != NOMATCH);
  
  ulTest = ulParserSemanticAnalysis(vpParserCtx, (void*)spData);
  GENASSERT(ulTest);
  
  // release the lexical analyzer parser
  vParserDtor(vpParserCtx);
  vpParserCtx = NULL;
  
  // test for lexer errors
  ulTest = ulGrammarErrorsCount(spData->vpErrorsCtx);
  GENASSERT(ulTest == 0);
  if(spData->sConfig.ulDisplayTokens)
  {
    vPrintTokens(spData);
  }
  
  //////////////////////////////////////////////////////////////////////////
  // parse the grammar tokens and translate to intermediate opcodes
  //////////////////////////////////////////////////////////////////////////
  // initialize - set the initial line number (one-based)
  spData->ulLineNo = 1;
  spData->ulRuleCount = ulVecSize(spData->vpIntRuleCtx);
  
  // create the intermediate opcode vector
  spData->vpIntOpcodesCtx = vpVecCtor(spData->vpMemCtx, sizeof(OP_INT_ALL), 1024);
  GENASSERT(spData->vpIntOpcodesCtx);
  
  ulTest = ulParserInit(vpParser, spData->vpMemCtx);
  GENASSERT(ulTest);
  vpParserCtx = vpParserCtor(vpParser, spData->vpMemCtx);
  GENASSERT(vpParserCtx);
  
  ucpTokIdx = (uchar*)vpVecFront(spData->vpTokIdxCtx);
  ulTokIdxSize = ulVecSize(spData->vpTokIdxCtx);
  sSrc.ucpStr = (uchar*)ucpTokIdx;
  sSrc.ulLen = ulTokIdxSize;
  ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, sSrc.ucpStr, sSrc.ulLen, NULL, APG_FALSE);
  GENASSERT(ulTest);
  ulParserState(vpParserCtx, &sState);
  if(spData->sConfig.ulDisplayState)
  {
    vPrintState(&sState, "*** Grammar Parser");
  }
  GENASSERT(sState.ulState != NOMATCH);
  
  ulTest = ulParserSemanticAnalysis(vpParserCtx, (void*)spData);
  GENASSERT(ulTest);
  vParserDtor(vpParserCtx);
  vpParserCtx = NULL;

  // test for parser errors
  ulTest = ulGrammarErrorsCount(spData->vpErrorsCtx);
  GENASSERT(ulTest == 0);
  
  // add syntactic and semantic controls to rule names
  vGrammarRuleNameControl(spData, &spData->sSyntaxControl, &spData->sSemanticsControl);
  
  // clean up unneeded token data
  if(spData->vpTokIdxCtx)
  {
    vMemFree(spData->vpMemCtx, spData->vpTokIdxCtx);
    spData->vpTokIdxCtx = NULL;
  }
  if(spData->vpTokTblCtx)
  {
    vMemFree(spData->vpMemCtx, spData->vpTokTblCtx);
    spData->vpTokTblCtx = NULL;
  }
  
  // remove redundant opcodes
  ulTest = ulReducedOpcodes(spData);
  if(ulDisplayOpcodes)
  {
    vPrintIntermediateOpcodes(spData);
  }
  
  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

// remove the redundant ALT(1), CAT(1) and REP(1,1) opcodes
typedef struct 
{
  ulong ulOpOffset;
  ulong ulLastIndex;
} INDEX_STACK;

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Remove redundant opcodes
//              ALT and CAT with only a single child
//              REP(n,m) with n == m == 1
//
// FUNCTION: static ulong ulReducedOpcodes
//
//      ARG: 1) GENCTX* spData
//              : pointer to a valid generator component context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulReducedOpcodes(GENCTX* spData)
{
  ulong ulReturn = APG_FALSE;
  void* vpOpcodes;
  void* vpStack = NULL;
  ulong ulIntIndex;
  INDEX_STACK* spStackFrame;
  ulong ulCount;
  ulong ulRuleCount = spData->ulRuleCount;
  ulong ulProdOffset;
  ulong ulProdCount;
  INT_RULE_LIST* spRule;
  INT_RULE_LIST* spEnd;
  OP_INT_ALL* spOp;
  OP_INT_ALL* spOpEnd;
  
  while(APG_TRUE)
  {
    spRule = (INT_RULE_LIST*)vpVecFront(spData->vpIntRuleCtx);
    spEnd = spRule + spData->ulRuleCount;
    spOp = (OP_INT_ALL*)vpVecFront(spData->vpIntOpcodesCtx);
    
    // count the opcodes
    ulCount = ulVecSize(spData->vpIntOpcodesCtx);
    
    // allocate a stack
    vpStack = vpVecCtor(spData->vpMemCtx, sizeof(INDEX_STACK), ulCount);
    GENASSERT(vpStack);
    
    // allocate the single opcode vector
    vpOpcodes = vpVecCtor(spData->vpMemCtx, sizeof(OP_INT_ALL), ulCount + 5);
    GENASSERT(vpOpcodes);
    ulIntIndex = 0;
    for(; spRule < spEnd; ++spRule)
    {
      spOp = (OP_INT_ALL*)vpVecAt(spData->vpIntOpcodesCtx, spRule->ulOpcodeOffset);
      spOpEnd = spOp + spRule->ulOpcodeCount;
      ulProdOffset = ulVecSize(vpOpcodes);
      ulProdCount = 0;
      for(; spOp < spOpEnd; ++spOp, ++ulIntIndex)
      {
        // skip unneeded opcodes
        if(spOp->ulOpId == OP_INT_ID_ALT && spOp->ulCount == 1){continue;}
        if(spOp->ulOpId == OP_INT_ID_CAT && spOp->ulCount == 1){continue;}
        if(spOp->ulOpId == OP_INT_ID_REP && spOp->ulMin == 1 && spOp->ulMax == 1){continue;}
        ++ulProdCount;
        
        // push a stack frame
        spStackFrame = (INDEX_STACK*)vpVecPush(vpStack, NULL);
        spStackFrame->ulOpOffset = ulVecSize(vpOpcodes);
        spStackFrame->ulLastIndex = spOp->ulIndexLast;
        
        // push an opcode
        memset((void*)&spOp->sPPTAttr, 0, sizeof(spOp->sPPTAttr));
        spOp->ulIndexThis = ulVecSize(vpOpcodes);
        vpVecPush(vpOpcodes, (void*)spOp);
        
        while(spStackFrame && spStackFrame->ulLastIndex == ulIntIndex)
        {
          OP_INT_ALL* spOpFinish;
          spStackFrame = (INDEX_STACK*)vpVecPop(vpStack);
          spOpFinish = (OP_INT_ALL*)vpVecAt(vpOpcodes, spStackFrame->ulOpOffset);
          spOpFinish->ulIndexLast = ulVecSize(vpOpcodes) - 1;
          spStackFrame = (INDEX_STACK*)vpVecBack(vpStack);
        }
      }
      
      spRule->ulOpcodeOffset = ulProdOffset;
      spRule->ulOpcodeCount = ulProdCount;
    }
    
    // swap intermediate opcode vectors
    vVecDtor(spData->vpIntOpcodesCtx);
    spData->vpIntOpcodesCtx = vpOpcodes;
    
    // success
    ulReturn = APG_TRUE;
    break;
  }
  
APG_EXIT:
  return ulReturn;
}

