/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/25/2007 12:57
  filename: D:\Projects\Apgv5_0\Generator\main.c
   purpose: Driver for the APG version 5.0 parser generator.
  
*********************************************************************/
#include "Generator.h"

#undef  __USE_PPT_VERSION
// remove or comment out the following #define statement to use the
// non-PPT version of the SABNF parsing components
#define __USE_PPT_VERSION

#define MAINASSERT(cond, msg) if(!(int)(cond)){vPrintLoc((char*)msg, __LINE__, __FILE__);break;}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: The driver (main) program for Apg.exe.
//
// FUNCTION: int main
//
//      ARG: 1) int argc
//              : number of command line arguments
//
//      ARG: 2) char** argv
//              : pointer to a list of pointers to the command line arguments
//
//   RETURN: 1 on success, 0 on failure
//
//////////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
  int       iReturn = 0;
  FILE*     spRedirectOut = NULL;
  GENCTX*   spGenCtx = NULL;
  CM_CONFIG sConfig;
  ulong     ulState;
  ulong     ulTest;
  double    dStartTime;
  double    dEndTime;
  double    dDuration;
  double    dBTest;
  double    dETest;
  double    dDTest;
  void*     vpMainLexerInit = NULL;
  void*     vpMainParserInit = NULL;

#ifdef __USE_PPT_VERSION
  // use these definitions to use the PPT version of the parsers
#include "PPTLexer.h"
#include "PPTParser.h"
  vpMainLexerInit = vpPPTLexerInit;
  vpMainParserInit = vpPPTParserInit;
#else
  // use these definitions to use the non-PPT version of the parsers
#include "Lexer.h"
#include "Parser.h"
  vpMainLexerInit = vpLexerInit;
  vpMainParserInit = vpParserInit;
#endif
  
  // single point of exit loop
  while(APG_TRUE)
  {
    // get the start clock time
    dStartTime = (double)clock();
  
    // set the alert handler to print location and terminate the application
    pfnSetAlertHandler(vTerminalAlert);
  
    // parse the command line
    ulState = ulCommandLine(&sConfig, argc, argv);
    MAINASSERT(ulState, "main: command line error detected");
  
    // print license notice to console
    vLicenseNotice();
    if(sConfig.ulLog)
    {
      // redirect output
      printf("main: %s - log file\n", sConfig.caLog);
      spRedirectOut = freopen(sConfig.caLog, "w", stdout);
      MAINASSERT(spRedirectOut, "main: unable to open redirect file");

      // print license notice to output file if different from console
      vLicenseNotice();
    }
  
    // constructor - initialize the generator component
    spGenCtx = spGenCtor(&sConfig);
    if(!spGenCtx){break;}
  
    // get and validate the grammar file
    ulTest = ulGrammar(spGenCtx);
    if(!ulTest){break;}
  
    // generate intermediate opcodes
    dBTest = (double)clock();
    ulTest = ulGrammarParser(spGenCtx, vpMainLexerInit, vpMainParserInit);
    dETest = (double)clock();
    dDTest = (dETest - dBTest) / (double)CLOCKS_PER_SEC;
    printf("%1.4e = Grammar Parsing Time(sec)\n", dDTest);
    if(!ulTest){break;}
  
    // collect the grammar metrics
    ulTest = ulGrammarMetrics(spGenCtx);
    if(!ulTest){break;}
  
    // collect the SEST metrics
    dBTest = (double)clock();
    ulTest = ulSESTMetrics(spGenCtx);
    dETest = (double)clock();
    dDTest = (dETest - dBTest) / (double)CLOCKS_PER_SEC;
    printf("%1.4e = SEST Metrics Time(sec)\n", dDTest);
    if(!ulTest){break;}
  
    // recursion analysis of the grammar rules
    if(!spGenCtx->sConfig.ulNoRecursionAnalysis)
    {
      dBTest = (double)clock();
      ulTest = ulSESTRecursionAnalysis(spGenCtx);
      dETest = (double)clock();
      dDTest = (dETest - dBTest) / (double)CLOCKS_PER_SEC;
      printf("%1.4e = Recursion Analysis Time(sec)\n", dDTest);
      if(!ulTest){break;}
    }
  
    // for printing selected branches of the SEST
    //       ulTest = ulBranchSESTMetrics(spGenCtx);
    //       if(!ulTest){break;}
    //       break;
  
    if(sConfig.ulPPT)
    {
      // generate PPTs
      dBTest = (double)clock();
      ulTest = ulPPTProductionOpcodes(spGenCtx);
      dETest = (double)clock();
      dDTest = (dETest - dBTest) / (double)CLOCKS_PER_SEC;
      printf("%1.4e = PPT Opcodes Time(sec)\n", dDTest);
      MAINASSERT(ulTest, "main: ulPPTProductionOpcodes() failed");
    
    }
    else
    {
      // generate production opcodes
      ulTest = ulProductionOpcodes(spGenCtx);
      if(!ulTest){break;}
    }
  
    if(spGenCtx->sConfig.ulParserName)
    {
      // output source code
      ulTest = ulGenOutput(spGenCtx, &spGenCtx->sConfig.caParserName[0]);
      MAINASSERT(ulTest, "main: ulGenOutput() failed");
    }
  
    // success
    iReturn = 1;
    break;
  }
  
  // get the generation time
  printf("\n");
  printf("*** Generation Time\n");
  dEndTime = (double)clock();
  dDuration = (dEndTime - dStartTime) / (double)CLOCKS_PER_SEC;
  printf("%1.4e = Time(sec)\n", dDuration);
  
  if(spGenCtx && spGenCtx->vpErrorsCtx && spGenCtx->vpWarningsCtx)
  {
    // display errors and warnings
    vGenErrors(spGenCtx);
  }

  // clean up resources and exit
  if(spGenCtx)
  {
    // display memory statistics
    vPrintMemStats(spGenCtx->vpMemCtx);
    
    // destructor - release ALL resources
    vGenDtor(spGenCtx);
  }
  
  if(spRedirectOut){fclose(spRedirectOut);}
  return iReturn;
}
