/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/30/2007 8:24
  filename: D:\Projects\Apgv5_0\Samples\MEGACO\Messages.c
   purpose: Separate a file of semicolon-separated MEGACO messages
            into a vector of BStrings.
            Timing and statistics tests on MEGACO messages.
  
*********************************************************************/
#include "main.h"

#define PREASSERT(cond) if(!(int)(cond)){printf("\nulParseMessage: [%lu] line number, [%lu] message error, message ignored\n", __LINE__, ulIndex);continue;}
static void vPrintStatistics(APG_PARSER_STATS* spStats);
static char* cpLookForStart(char* cpBeg, char* cpEnd);
static char* cpLookForEnd(char* cpBeg, char* cpEnd);

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Separate a file of semicolon-separated MEGACO messages
//              into a vector of BStrings.
//
// FUNCTION: ulong ulMessages
//
//      ARG: 1) char* cpFile
//              : pointer to the (in memory) file of messages
//
//      ARG: 2) ulong ulLen
//              : length of the file
//
//      ARG: 3) void* vpBstr
//              : pointer to a valid vector of binary strings component context
//
//   RETURN: number of messages in the file
//
//////////////////////////////////////////////////////////////////////////
ulong ulMessages(char* cpFile, ulong ulLen, void* vpBstr)
{
  ulong ulReturn;
  ulong ulCount;
  ulong ulIndex;
  APG_BSTR* spMsg;
  char* cpStart;
  char* cpEnd;
  char* cpNext;
  char* cpEndFile;
  DASSERT(cpFile);
  DASSERT(ulLen);
  DASSERT(vpBstr);

  cpNext = cpFile;
  cpEndFile = cpNext + ulLen;
  // single-point-of-exit loop
  while(APG_TRUE)
  {
    cpStart = cpLookForStart(cpNext, cpEndFile);
    if(!cpStart){break;}
  
    cpEnd = cpLookForEnd(cpStart, cpEndFile);
    if(!cpEnd){break;}

    // push a message on the vector
    spMsg = (APG_BSTR*)vpVecPush(vpBstr, NULL);
    if(!cpEnd){goto APG_EXIT;}
    spMsg->ucpStr = (uchar*)cpStart;
    spMsg->ulLen = (ulong)(cpEnd - cpStart);

    cpNext = cpEnd;
  } // end of single-point-of-exit loop

  // null-term each message
  // - OK since this character must be semicolon
  // - OK because this is a temporary copy of the file
  ulCount = ulVecSize(vpBstr);
  MASSERT(ulCount);
  spMsg = (APG_BSTR*)vpVecFront(vpBstr);
  MASSERT(spMsg);
  for(ulIndex = 0; ulIndex < ulCount; ++ulIndex, ++spMsg)
  {
    cpEnd = (char*)spMsg->ucpStr + spMsg->ulLen;
    *cpEnd = 0;
  }

  // success
  ulReturn = ulCount;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Search the file for the start of the next MEGACO message
//
// FUNCTION: static char* cpLookForStart
//
//      ARG: 1) char* cpBeg
//              : character in the file to begin the search from
//
//      ARG: 2) char* cpEnd
//              : pointer to the end of the file
//
//   RETURN: on success, pointer to the beginning of a MEGACO message
//           on failure, NULL
//
//////////////////////////////////////////////////////////////////////////
static char* cpLookForStart(char* cpBeg, char* cpEnd)
{
  char* cpReturn = NULL;
  char cPrev = LF;
  char* cpNext = cpBeg;

  for(; cpNext < cpEnd; ++cpNext)
  {
    if(cPrev == LF && *cpNext != SEMI)
    {
      // found begin of msg
      break;
    }

    cPrev = *cpNext;
  }

  if(cpNext < cpEnd)
  {
    // found a msg
    cpReturn = cpNext;
  }
  // else found end of text

  return cpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: 
//
// FUNCTION: static char* cpLookForEnd
//
//      ARG: 1) char* cpBeg
//              : pointer to the beginning of a MEGACO message
//
//      ARG: 2) char* cpEnd
//              : pointer to the end of the file
//
//   RETURN: on success, pointer to the end of a MEGACO message
//           on failure, NULL
//
//////////////////////////////////////////////////////////////////////////
static char* cpLookForEnd(char* cpBeg, char* cpEnd)
{
  char* cpReturn = NULL;
  char cPrev = LF;
  char* cpNext = cpBeg;
  
  for(; cpNext < cpEnd; ++cpNext)
  {
    if(cPrev == LF && *cpNext == SEMI)
    {
      // found end of msg
      break;
    }
    
    cPrev = *cpNext;
  }
  
  if(cpBeg < cpNext)
  {
    // found a msg
    cpReturn = cpNext;
  }
  // else found end of text
  
  return cpReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Parses all of the messages in a list of messages and
//              reports on the collective parsing statistics.
//
// FUNCTION: ulong ulStatisticsTest
//
//      ARG: 1) void* vpParserInit
//              : pointer to valid APG parser data
//                (as defined in a parser header file)
//
//      ARG: 2) void* vpMsgs
//              : pointer to a valid vector of binary strings context
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulStatisticsTest(void* vpParserInit, void* vpMsgs, APG_PARSER_STATS* spTotalStats)
{
  ulong         ulReturn = APG_FALSE;
  APG_BSTR*     spMsg;
  ulong         ulIndex;
  ulong         ulCount;
  ulong         ulTest;
  ulong         ulTrace = APG_FALSE;
  void*         vpData = NULL;
  void*         vpParserCtx;
  APG_PARSER_STATS  sStats;
  ulong ulSize = sizeof(APG_PARSER_STATS) / sizeof(ulong);
  APG_PARSER_STATE  sState;
  DASSERT(vpMsgs);
  
  // get a parser context
  vpParserCtx = vpParserCtor(vpParserInit, vpMemCtx);
  MASSERT(vpParserCtx);
  
  // loop over all msgs
  memset((void*)&sStats, 0, sizeof(APG_PARSER_STATS));
  memset((void*)spTotalStats, 0, sizeof(APG_PARSER_STATS));
  spMsg = (APG_BSTR*)vpVecFront(vpMsgs);
  MASSERT(spMsg);
  ulCount = ulVecSize(vpMsgs);
  MASSERT(ulCount);
  for(ulIndex = 0; ulIndex < ulCount; ++ulIndex, ++spMsg)
  {
//     if(ulIndex != 27){continue;}
//     ulTrace = APG_TRUE;
    ulTest = ulParserSyntaxAnalysis(vpParserCtx, 0, spMsg->ucpStr, spMsg->ulLen, vpData, ulTrace);
    PREASSERT(ulTest);
    ulTest = ulParserState(vpParserCtx, &sState);
    PREASSERT(ulTest);
    if(sState.ulState == NOMATCH)
    {
      printf("\n");
      printf("*** PARSE FAILED FOR THIS MESSAGE ***\n");
      printf("%s\n", (char*)spMsg->ucpStr);
      PREASSERT(APG_FALSE);
    }
    
    if(sState.ulState == MATCH)
    {
      ulTest = ulParserStats(vpParserCtx, &sStats);
      PREASSERT(ulTest);
      for(ulTest = 0; ulTest < ulSize; ++ulTest)
      {
        ulong* ulpState = (ulong*)&sStats;
        ulong* ulpTotal = (ulong*)spTotalStats;
        ulpTotal[ulTest] += ulpState[ulTest];
      }
    }
    
    // clear the error message queue
    ulTest = ulErrCount(vpMemCtx);
    if(ulTest)
    {
      vPrintErrors(vpMemCtx);
      vErrClear(vpMemCtx);
      PREASSERT(APG_FALSE);
    }
  }
  
  // print the total stats
  vPrintStatistics(spTotalStats);
  
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Compares timing results between APG parsers with
//              and without PPTs.
//
// FUNCTION: ulong ulTimingTest
//
//      ARG: 1) void* vpMsgs
//              : pointer to a valid vector of binary strings context
//
//      ARG: 2) ulong ulRepetitions
//              : number of times to parse each message in the list
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
ulong ulTimingTest(void* vpMsgs, ulong ulRepetitions)
{
  APG_PARSER_STATE  sState;
  ulong     ulReturn = APG_FALSE;
  void*     vpCtx = NULL;
  void*     vpPPTCtx = NULL;
  APG_BSTR* spMsgs;
  ulong     ulTest;
  ulong     ulCount;
  ulong     ulAnn, ulRep;
  ulong     ulMsgs;
  ulong     ulMsgChars;
  double    dStart;
  double    dFinish;
  double    dClocksPerSec;
  double    dDuration;
  double    dDurationPPT;
  double    dMsgPerMsec;
  double    dCharsPerMsec;
  
  // get a parser components
  vpCtx = vpParserCtor(vpMEGACOInit, vpMemCtx);
  MASSERT(vpCtx);
  vpPPTCtx = vpParserCtor(vpMEGACOPPTInit, vpMemCtx);
  MASSERT(vpPPTCtx);
  
  spMsgs = (APG_BSTR*)vpVecFront(vpMsgs);
  MASSERT(spMsgs);
  ulCount = ulVecSize(vpMsgs);
  MASSERT(ulCount);

  //////////////////////////////////////////////////////////////////////////
  // MEGACO without PPT test
  //////////////////////////////////////////////////////////////////////////
  // iterate through repetitions
  ulMsgs = 0;
  ulMsgChars = 0;
  dClocksPerSec = (double)CLOCKS_PER_SEC;
  dStart = (double)clock();
  for(ulRep = 0; ulRep < ulRepetitions; ++ulRep)
  {
    // iterate through the announcements
    spMsgs = (APG_BSTR*)vpVecFront(vpMsgs);
    MASSERT(spMsgs);
    for(ulAnn = 0; ulAnn < ulCount; ++ulAnn, ++spMsgs)
    {
      ++ulMsgs;
      ulMsgChars += spMsgs->ulLen;
      ulTest = ulParserSyntaxAnalysis(vpCtx, 0, spMsgs->ucpStr, spMsgs->ulLen, NULL, APG_FALSE);
      MASSERT(ulTest);
      ulTest = ulParserState(vpCtx, &sState);
      MASSERT(ulTest);
      MASSERT(sState.ulState != NOMATCH);
    }
  }

  // compute rates
  dFinish = (double)clock();
  dDuration = ((dFinish - dStart) * 1000.0) / dClocksPerSec;
  dMsgPerMsec = (double)ulMsgs / dDuration;
  dCharsPerMsec = (double)ulMsgChars / dDuration;
  
  printf("MEGACO without PPT\n");
  printf(" - Messages = %lu\n", ulMsgs);
  printf(" - Characters = %lu\n", ulMsgChars);
  printf(" - Time(msec) = %1.4g\n", dDuration);
  printf(" - Messages/msec = %1.4g\n", dMsgPerMsec);
  printf(" - Characters/msec = %1.4g\n", dCharsPerMsec);

  //////////////////////////////////////////////////////////////////////////
  // MEGACO without PPT test
  //////////////////////////////////////////////////////////////////////////
  // iterate through repetitions
  ulMsgs = 0;
  ulMsgChars = 0;
  dClocksPerSec = (double)CLOCKS_PER_SEC;
  dStart = (double)clock();
  for(ulRep = 0; ulRep < ulRepetitions; ++ulRep)
  {
    // iterate through the announcements
    spMsgs = (APG_BSTR*)vpVecFront(vpMsgs);
    MASSERT(spMsgs);
    for(ulAnn = 0; ulAnn < ulCount; ++ulAnn, ++spMsgs)
    {
      ++ulMsgs;
      ulMsgChars += spMsgs->ulLen;
      ulTest = ulParserSyntaxAnalysis(vpPPTCtx, 0, spMsgs->ucpStr, spMsgs->ulLen, NULL, APG_FALSE);
      MASSERT(ulTest);
      ulTest = ulParserState(vpCtx, &sState);
      MASSERT(ulTest);
      MASSERT(sState.ulState != NOMATCH);
    }
  }
  
  // compute rates
  dFinish = (double)clock();
  dDurationPPT = ((dFinish - dStart) * 1000.0) / dClocksPerSec;
  dMsgPerMsec = (double)ulMsgs / dDurationPPT;
  dCharsPerMsec = (double)ulMsgChars / dDurationPPT;
 
  printf("\n");
  printf("MEGACO with PPT\n");
  printf(" - Messages = %lu\n", ulMsgs);
  printf(" - Characters = %lu\n", ulMsgChars);
  printf(" - Time(msec) = %1.4g\n", dDurationPPT);
  printf(" - Messages/msec = %1.4g\n", dMsgPerMsec);
  printf(" - Characters/msec = %1.4g\n", dCharsPerMsec);

  // compute PPT factor
  printf("\n");
  printf(" -         PPT factor = %1.4g\n", dDuration/dDurationPPT);

  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:

  // clean up
  if(vpCtx){vParserDtor(vpCtx);}
  if(vpPPTCtx){vParserDtor(vpPPTCtx);}
  return ulReturn;
}

// print a parser statistics
//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Helper function for printing parser statistics.
//              Prints statistics for a specific operator type (ALT, CAT etc.).
//
// FUNCTION: static void vPrintOperatorLine
//
//      ARG: 1) char* cpOp
//              : name of the operator type
//
//      ARG: 2) APG_NODE_STATS* spNode
//              : statistics struct for the operator type
//
//      ARG: 3) APG_NODE_STATS* spTotal
//              : pointer to a statistics struct to collect totals in
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vPrintOperatorLine(char* cpOp, APG_NODE_STATS* spNode, APG_NODE_STATS* spTotal)
{
  DASSERT(cpOp && spNode);
  
  if(spTotal)
  {
    spTotal->ulVisits += spNode->ulVisits;
    spTotal->ulEmpty += spNode->ulEmpty;
    spTotal->ulMatched += spNode->ulMatched;
    spTotal->ulNotMatched += spNode->ulNotMatched;
  }
  printf("%5s %7lu %7lu %7lu %7lu",
    cpOp,
    spNode->ulVisits,
    spNode->ulMatched,
    spNode->ulEmpty,
    spNode->ulNotMatched);
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Prints APG parsing statistics
//
// FUNCTION: static void vPrintStatistics
//
//      ARG: 1) APG_PARSER_STATS* spStats
//              : pointer to a valid APG parser statistics struct.
//
//   RETURN: void
//
//////////////////////////////////////////////////////////////////////////
static void vPrintStatistics(APG_PARSER_STATS* spStats)
{
  ulong           ulTemp;
  APG_NODE_STATS* spNode;
  APG_NODE_STATS  sNodeTotal;
  DASSERT(spStats);
  
  printf("\n");
  printf("*************************\n");
  printf("*** Parser Statistics ***\n");
  printf("*************************\n");
  printf("\n");
  printf("BACK TRACKING\n");
  printf("back track occurrences = %lu\n", spStats->ulBackTrackCount);
  printf("back tracked characters= %lu\n", spStats->ulBackTrackChars);
  
  printf("\n");
  printf("OPERATORS TESTED\n");
  printf("%5s %7s %7s %7s %7s %7s\n", "", "total", "MATCH", "EMPTY", "NOMATCH", "CHILDREN");
  memset((void*)&sNodeTotal, 0, sizeof(sNodeTotal));
  vPrintOperatorLine("RNM", &spStats->sRnm, &sNodeTotal);
  printf("\n");
  vPrintOperatorLine("ALT", &spStats->sAlt, &sNodeTotal);
  printf(" %7lu\n", spStats->sAlt.ulChildCount);
  vPrintOperatorLine("CAT", &spStats->sCat, &sNodeTotal);
  printf(" %7lu\n", spStats->sCat.ulChildCount);
  vPrintOperatorLine("REP", &spStats->sRep, &sNodeTotal);
  printf(" %7lu\n", spStats->sRep.ulChildCount);
  vPrintOperatorLine("PRD", &spStats->sPrd, &sNodeTotal);
  printf("\n");
  vPrintOperatorLine("TRG", &spStats->sTrg, &sNodeTotal);
  printf("\n");
  vPrintOperatorLine("TBS", &spStats->sTbs, &sNodeTotal);
  printf("\n");
  vPrintOperatorLine("TLS", &spStats->sTls, &sNodeTotal);
  printf("\n");
  vPrintOperatorLine("TOTAL", &sNodeTotal, NULL);
  ulTemp = spStats->sAlt.ulChildCount + spStats->sCat.ulChildCount + spStats->sRep.ulChildCount;
  printf(" %7lu\n", ulTemp);
  
  printf("\n");
  printf("PREDICTIVE PARSING TABLES TESTED\n");
  printf("%5s %7s %7s %7s %7s %7s\n", "", "total", "MATCH", "EMPTY", "NOMATCH", "ACTIVE");
  spNode = &spStats->sPpt;
  printf("%5s %7lu %7lu %7lu %7lu %7lu\n",
    "PPT",
    spNode->ulVisits,
    spNode->ulMatched,
    spNode->ulEmpty,
    spNode->ulNotMatched,
    spNode->ulPossibleMatched);
  
  printf("\n");
  printf("EOF CHARACTERS TESTED\n");
  printf("%5s %7s %7s\n", "", "total", "MATCH");
  spNode = &spStats->sEof;
  printf("%5s %7lu %7lu\n",
    "EOF",
    spNode->ulVisits,
    spNode->ulMatched);
}
