/********************************************************************
  APG - an ABNF Parser Generator
  Copyright  (C) 2007 Coast to Coast Research, Inc.
 
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see
  <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>
  or write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

    author: Lowell Thomas
            lowell@coasttocoastresearch.com
            http://www.coasttocoastresearch.com

   created: 8/28/2007 9:37
  filename: D:\Projects\Apgv5_0\Samples\SIP\TortureTests.c
   purpose: Translates the "long lines", "non-printable characters"
                and "long repeating strings" from the torture test SIP
                messages given in RFC 4475.

                Input file is printable ASCII characters only. Output
                is a binary file containing non-printable characters,
                very long lines, and lines containing very long
                strings of repeating characters.
                Both input and output files contain multiple SIP
                messages, separated by one or more lines beginning
                with a semicolon - that is a "CRLF;" character sequence.
                Note that no valid SIP message line begins with "CRLF;"
  
*********************************************************************/
#include "main.h"

static ulong ulLongLines(APG_BSTR* spMsg, void* vpTransCtx);
static ulong ulBinaryCharacters(APG_BSTR* spMsg, void* vpTransCtx);
static ulong ulRepeatingStrings(APG_BSTR* spMsg, void* vpTransCtx);

static char* g_cpLineEnd = "\r\n";
// quick convert from hex digit to integer
// quick character validation
static uchar ucaHexToInt[256] = 
{
  255,255,255,255,255,255,255,255,255,255,  // 0
  255,255,255,255,255,255,255,255,255,255,  // 10
  255,255,255,255,255,255,255,255,255,255,  // 20
  255,255,255,255,255,255,255,255,255,255,  // 30
  255,255,255,255,255,255,255,255,  0,  1,  // 40
    2,  3,  4,  5,  6,  7,  8,  9,255,255,  // 50
  255,255,255,255,255, 10, 11, 12, 13, 14,  // 60
   15,255,255,255,255,255,255,255,255,255,  // 70
  255,255,255,255,255,255,255,255,255,255,  // 80
  255,255,255,255,255,255,255, 10, 11, 12,  // 90
   13, 14, 15,255,255,255,255,255,255,255,  // 100
  255,255,255,255,255,255,255,255,255,255,  // 110
  255,255,255,255,255,255,255,255,255,255,  // 120
  255,255,255,255,255,255,255,255,255,255,  // 130
  255,255,255,255,255,255,255,255,255,255,  // 140
  255,255,255,255,255,255,255,255,255,255,  // 150
  255,255,255,255,255,255,255,255,255,255,  // 160
  255,255,255,255,255,255,255,255,255,255,  // 170
  255,255,255,255,255,255,255,255,255,255,  // 180
  255,255,255,255,255,255,255,255,255,255,  // 190
  255,255,255,255,255,255,255,255,255,255,  // 200
  255,255,255,255,255,255,255,255,255,255,  // 210
  255,255,255,255,255,255,255,255,255,255,  // 220
  255,255,255,255,255,255,255,255,255,255,  // 230
  255,255,255,255,255,255,255,255,255,255,  // 240
  255,255,255,255,255,255                   // 250
};

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: 
//    Translates the "long lines", "non-printable characters"
//    and "long repeating strings" from the torture test SIP
//    messages given in RFC 4475.
// 
// Input file is printable ASCII characters only. Output
// is a binary file containing non-printable characters,
// very long lines, and lines containing very long
// strings of repeating characters.
// Both input and output files contain multiple SIP
// messages, separated by one or more lines beginning
// with a semicolon - that is a "CRLF;" character sequence.
// Note that no valid SIP message line begins with "CRLF;"
//
// FUNCTION: ulong ulTortureTestTranslator
//
//      ARG: 1) char* cpInput
//              : Input file is printable ASCII characters only
//
//      ARG: 2) char* cpOutput
//               : Input file is printable ASCII characters only. Output
//                 is a binary file containing non-printable characters,
//                 very long lines, and lines containing very long
//                 strings of repeating characters.
//
//   RETURN: 
// returns the number of SIP messages in the output file.
// returns zero if an error is encountered.
//
//////////////////////////////////////////////////////////////////////////
ulong ulTortureTestTranslator(char* cpInput, char* cpOutput)
{
  ulong ulReturn = 0;
  APG_BSTR  sSrc;
  APG_BSTR  sTemp;
  void*     vpMsgsInCtx = NULL;
  void*     vpMsgsOutCtx = NULL;
  void*     vpTransOneCtx = NULL;
  void*     vpTransTwoCtx = NULL;
  FILE*     spFile = NULL;
  ulong     ulBytesWritten = 0;
  ulong     ulBytes;
  DASSERT(cpInput);
  DASSERT(cpInput[0] != 0);
  DASSERT(cpOutput);
  DASSERT(cpOutput[0] != 0);

  // get the list of messages
  sSrc.ulLen = ulGetFileSize(cpInput);
  SIPASSERT(sSrc.ulLen);
  sSrc.ucpStr = (uchar*)vpMemAlloc(vpMemCtx, sSrc.ulLen + 10);
  SIPASSERT(sSrc.ucpStr);
  sSrc.ulLen = ulGetFile(cpInput, (void*)sSrc.ucpStr);
  SIPASSERT(sSrc.ulLen);
  sSrc.ucpStr[sSrc.ulLen] = 0;
  
  // get a vector contexts
  vpMsgsInCtx = vpVecCtor(vpMemCtx, sizeof(APG_BSTR), 0);
  SIPASSERT(vpMsgsInCtx);
  vpMsgsOutCtx = vpVecCtor(vpMemCtx, sizeof(APG_BSTR), 0);
  SIPASSERT(vpMsgsOutCtx);
  vpTransOneCtx = vpVecCtor(vpMemCtx, sizeof(uchar), 0);
  SIPASSERT(vpTransOneCtx);
  vpTransTwoCtx = vpVecCtor(vpMemCtx, sizeof(uchar), 0);
  SIPASSERT(vpTransTwoCtx);

  // open the output file
  spFile = fopen(cpOutput, "wb");
  SIPASSERT(spFile);
  
  // translate the long lines
  ulBytes = ulLongLines(&sSrc, vpTransOneCtx);
  SIPASSERT(ulBytes);
  
  // translate the long repeating strings
  sTemp.ucpStr = (uchar*)vpVecFront(vpTransOneCtx);
  SIPASSERT(sTemp.ucpStr);
  sTemp.ulLen = ulVecSize(vpTransOneCtx);
  ulBytes = ulRepeatingStrings(&sTemp, vpTransTwoCtx);
  SIPASSERT(ulBytes);
  
  // translate the non-printing characters
  sTemp.ucpStr = (uchar*)vpVecFront(vpTransTwoCtx);
  SIPASSERT(sTemp.ucpStr);
  sTemp.ulLen = ulVecSize(vpTransTwoCtx);
  vVecClear(vpTransOneCtx);
  ulBytes = ulBinaryCharacters(&sTemp, vpTransOneCtx);
  SIPASSERT(ulBytes);
  
  // write the translation to the output file
  sTemp.ucpStr = (uchar*)vpVecFront(vpTransOneCtx);
  SIPASSERT(sTemp.ucpStr);
  sTemp.ulLen = ulVecSize(vpTransOneCtx);
  ulBytesWritten = fwrite((void*)sTemp.ucpStr, 1, sTemp.ulLen, spFile);
  SIPASSERT(ulBytesWritten == sTemp.ulLen);

  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  if(spFile){fclose(spFile);}
  if(vpTransOneCtx){vVecDtor(vpTransOneCtx);}
  if(vpTransTwoCtx){vVecDtor(vpTransTwoCtx);}
  if(vpMsgsOutCtx){vVecDtor(vpMsgsOutCtx);}
  if(vpMsgsInCtx){vVecDtor(vpMsgsInCtx);}
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Search for tagged long lines and replace them as directed.
//
// FUNCTION: static ulong ulLongLines
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining a SIP message
//
//      ARG: 2) void* vpTransCtx
//              : vector component context for the output file
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulLongLines(APG_BSTR* spMsg, void* vpTransCtx)
{
  ulong ulReturn = APG_FALSE;
  ulong ulPrefixLen;
  char* cpPrefixEnd;
  char* cpPhrasePrev;
  char* cpPhraseNext;
  char* cpPhraseEnd;
  char* cpLineBeg;
  char* cpLineEnd;
  void* vpTest;
  char  cSave;
  static char* cpStartTag = "<allOneLine>\r\n";
  static char* cpEndTag = "</allOneLine>\r\n";
  ulong ulStartTagLen = strlen(cpStartTag);
  ulong ulEndTagLen = strlen(cpEndTag);

  // single-point-of-exit loop
  cpPhrasePrev = (char*)spMsg->ucpStr;
  cpPrefixEnd = cpPhrasePrev + spMsg->ulLen;
  while(APG_TRUE)
  {
    // find the next long line
    cpPhraseNext = strstr(cpPhrasePrev, cpStartTag);
    if(!cpPhraseNext){break;}
    cpPhraseEnd = strstr(cpPhraseNext, cpEndTag);
    SIPASSERT(cpPhraseEnd);

    // output the prefix, if any
    ulPrefixLen = (ulong)(cpPhraseNext - cpPhrasePrev);
    if(ulPrefixLen)
    {
      vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
      SIPASSERT(vpTest);
    }

    // remove the extra CRLFs
    cSave = *cpPhraseEnd;
    *cpPhraseEnd = 0;
    cpPhraseNext += ulStartTagLen;
    cpLineBeg = cpPhraseNext;
    while(APG_TRUE)
    {
      cpLineEnd = strstr(cpLineBeg, g_cpLineEnd);
      if(!cpLineEnd){break;}
      ulPrefixLen = (ulong)(cpLineEnd - cpLineBeg);
      if(ulPrefixLen)
      {
        vpTest = vpVecPushn(vpTransCtx, (void*)cpLineBeg, ulPrefixLen);
        SIPASSERT(vpTest);
      }
    
      cpLineBeg = cpLineEnd + strlen(g_cpLineEnd);
    } // end of single-point-of-exit loop
    *cpPhraseEnd = cSave;

    // output a line end
    vpTest = vpVecPushn(vpTransCtx, (void*)g_cpLineEnd, strlen(g_cpLineEnd));
    SIPASSERT(vpTest);
    
    // set up for next phrase 
    cpPhrasePrev = cpPhraseEnd + ulEndTagLen;
  } // end of single-point-of-exit loop

  // output the tail
  ulPrefixLen = (ulong)(cpPrefixEnd - cpPhrasePrev);
  if(ulPrefixLen)
  {
    vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
    SIPASSERT(vpTest);
  }

  // success
  ulReturn = APG_TRUE;

APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Search for tagged repeating strings and replace as directed.
//
// FUNCTION: static ulong ulRepeatingStrings
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining a SIP message
//
//      ARG: 2) void* vpTransCtx
//              : vector component context for the output file
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulRepeatingStrings(APG_BSTR* spMsg, void* vpTransCtx)
{
  ulong ulReturn = APG_FALSE;
  ulong ulPrefixLen;
  char* cpPrefixEnd;
  char* cpPhrasePrev;
  char* cpPhraseNext;
  char* cpPhraseEnd;
  void* vpTest;
  char  cSave;
  ulong ulCount;
  ulong ulIndex;
  char* cpRepeat;
  static char* cpStartTag = "<repeat count=";
  static char* cpEndTag = "</repeat>";
  ulong ulStartTagLen = strlen(cpStartTag);
  ulong ulEndTagLen = strlen(cpEndTag);
  
  // single-point-of-exit loop
  cpPhrasePrev = (char*)spMsg->ucpStr;
  cpPrefixEnd = cpPhrasePrev + spMsg->ulLen;
  while(APG_TRUE)
  {
    // find the next long line
    cpPhraseNext = strstr(cpPhrasePrev, cpStartTag);
    if(!cpPhraseNext){break;}
    cpPhraseEnd = strstr(cpPhraseNext, cpEndTag);
    SIPASSERT(cpPhraseEnd);
    
    // output the prefix, if any
    ulPrefixLen = (ulong)(cpPhraseNext - cpPhrasePrev);
    if(ulPrefixLen)
    {
      vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
      SIPASSERT(vpTest);
    }
    
    // find the count
    cpRepeat = strstr(cpPhraseNext, ">");
    SIPASSERT(cpRepeat);
    cSave = *cpRepeat;
    *cpRepeat = 0;
    ulCount = (ulong)atol(cpPhraseNext + ulStartTagLen);
    *cpRepeat = cSave;

    // find the repeat string
    ++cpRepeat;

    // write the repeated string
    cSave = *cpPhraseEnd;
    *cpPhraseEnd = 0;
    ulPrefixLen = (ulong)strlen(cpRepeat);
    for(ulIndex = 0; ulIndex < ulCount; ++ulIndex)
    {
      vpTest = vpVecPushn(vpTransCtx, (void*)cpRepeat, ulPrefixLen);
      SIPASSERT(vpTest);
    }
    *cpPhraseEnd = cSave;
    
    // set up for next phrase 
    cpPhrasePrev = cpPhraseEnd + ulEndTagLen;
  } // end of single-point-of-exit loop
  
  // output the tail
  ulPrefixLen = (ulong)(cpPrefixEnd - cpPhrasePrev);
  if(ulPrefixLen)
  {
    vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
    SIPASSERT(vpTest);
  }
  
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}

//////////////////////////////////////////////////////////////////////////
// DESCRIPTION: Search for tagged binary characters and replace as directed.
//
// FUNCTION: static ulong ulBinaryCharacters
//
//      ARG: 1) APG_BSTR* spMsg
//              : binary string defining a SIP message
//
//      ARG: 2) void* vpTransCtx
//              : vector component context for the output file
//
//   RETURN: true on success
//
//////////////////////////////////////////////////////////////////////////
static ulong ulBinaryCharacters(APG_BSTR* spMsg, void* vpTransCtx)
{
  ulong ulReturn = APG_FALSE;
  ulong ulPrefixLen;
  char* cpPrefixEnd;
  char* cpPhrasePrev;
  char* cpPhraseNext;
  char* cpPhraseEnd;
  void* vpTest;
  static char* cpStartTag = "<hex>";
  static char* cpEndTag = "</hex>";
  ulong ulStartTagLen = strlen(cpStartTag);
  ulong ulEndTagLen = strlen(cpEndTag);
  
  // single-point-of-exit loop
  cpPhrasePrev = (char*)spMsg->ucpStr;
  cpPrefixEnd = cpPhrasePrev + spMsg->ulLen;
  while(APG_TRUE)
  {
    // find the next binary phrase
    cpPhraseNext = strstr(cpPhrasePrev, cpStartTag);
    if(!cpPhraseNext){break;}
    cpPhraseEnd = strstr(cpPhraseNext, cpEndTag);
    SIPASSERT(cpPhraseEnd);
    
    // output the prefix, if any
    ulPrefixLen = (ulong)(cpPhraseNext - cpPhrasePrev);
    if(ulPrefixLen)
    {
      vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
      SIPASSERT(vpTest);
    }
    
    // translate all hex characters
    cpPhraseNext += ulStartTagLen;
    while(APG_TRUE)
    {
      uchar ucBinary;
      uchar ucDigit1;
      uchar ucDigit2;

      if(cpPhraseNext == cpPhraseEnd){break;}
      if(*cpPhraseNext == CR || *cpPhraseNext == LF)
      {
        // skip line enders
        ++cpPhraseNext;
        continue;
      }

      // convert digit one
      ucDigit1 = ucaHexToInt[(uchar)*cpPhraseNext];
      SIPASSERT(ucDigit1 != 255);
      
      // convert digit two
      ++cpPhraseNext;
      SIPASSERT(cpPhraseNext < cpPhraseEnd);
      ucDigit2 = ucaHexToInt[(uchar)*cpPhraseNext];
      SIPASSERT(ucDigit2 != 255);
      
      // write the binary value
      ucBinary = (ucDigit1 << 4) + ucDigit2;
      vpTest = vpVecPush(vpTransCtx, (void*)&ucBinary);
      SIPASSERT(vpTest);
      
      ++cpPhraseNext;
    } // end of single-point-of-exit loop
    
    // set up for next phrase 
    cpPhrasePrev = cpPhraseEnd + ulEndTagLen;
  } // end of single-point-of-exit loop
  
  // output the tail
  ulPrefixLen = (ulong)(cpPrefixEnd - cpPhrasePrev);
  if(ulPrefixLen)
  {
    vpTest = vpVecPushn(vpTransCtx, (void*)cpPhrasePrev, ulPrefixLen);
    SIPASSERT(vpTest);
  }
  
  // success
  ulReturn = APG_TRUE;
  
APG_EXIT:
  return ulReturn;
}
