/*! \file Agent.cpp
 <pre>
 <b>File:</b> Agent.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/10/2008
 <b>Contents:</b> Implementation of main agent class.
 </pre>
 */

#include "Agent.h"
#include "Config.h"
#include "Skills.h"
#include "TSocket.h"
#include "Logger.h"
#include "HSTest.h"
#include "HSAttack.h"

#ifdef WIN32
#   include <windows.h>
#endif

#include <iostream>

using namespace std;

extern Config *config;

Agent::Agent() {
    
    parser = new Parser();
    world = new WorldModel();
    player = new PlayerModel();
    
    actualLowSkill = NULL;
    actualMiddleSkill = NULL;
    actualHighSkill = NULL;
    
    running = true;
}

Agent::~Agent() {
    
    delete parser;
    delete world;
    delete player;
}

int Agent::runMultiThreaded() {
    
    Communication *comm = new Communication(config->serverIP, config->serverPort);
    
    comm->send("(scene rsg/agent/soccerbot056.rsg)");
    comm->receive();
    
    sprintf(message, "(init (unum %d)(teamname %s))", config->player.number, config->player.team);
    comm->send(message);
    comm->receive();
    
    comm->startReceiving();
    
    while (running) {
        
        //receive message from server
        comm->getMessage(message);
        
        //parse message
        bool parseTest = parser->parseData(message);
        message[0] = '\0';
        
        if (!parseTest) {
            
            continue;
        }
        
        //worldmodel update
        world->update(parser->apgparser.GetStoredData());
        
        //playermodel update
        player->update(parser->apgparser.GetStoredData());
        
        //player thinking
        if (world->playMode != PM_BeforeKickOff) {
            
            behave();
        }
        
        //add player model data to message
        player->buildMessage(message);
        
        //send message to server
        comm->send(message);
    }
    
    comm->stopReceiving();
    delete comm;
    return 0;
}

int Agent::stop() {
    
    running = false;
    return 0;
}

int Agent::runSingleThreaded() {
    
    TSocket *socket = new TSocket();
    if (!socket->connect(config->serverIP, config->serverPort)) {
        
        return 1;
    }
    
    socket->send("(scene rsg/agent/soccerbot056.rsg)");
    socket->receive(message, sizeof(message));
    
    sprintf(message, "(init (unum %d)(teamname %s))", config->player.number, config->player.team);
    socket->send(message);
    socket->receive(message, sizeof(message));
    
    while (running) {
        
        if (socket->select(0, 10000)) {
            
            //receive message
            if (!socket->receive(message, sizeof(message))) {
                
                socket->close();
                return 0;
            }
            
#ifdef _LOGGER
            Logger::logoutput(Logger::LOG_RECIEVED_DATA, log_recieved, message);
#endif
            
            //parse message
            bool parseTest = parser->parseData(message);
            message[0] = '\0';
            
            if (!parseTest) {
                
                continue;
            }
            
#ifdef _LOGGER
            Logger::logoutput(Logger::LOG_PARSE_TEST, log_parse_test, parse_test ? "ok" : "error");
#endif
            
            //worldmodel update
            world->update(parser->apgparser.GetStoredData());
            
            //playermodel update
            player->update(parser->apgparser.GetStoredData());
            
#ifdef _LOGGER
            player->logPlayerModel();
            Logger::logoutput(Logger::LOG_TEMP, log_simplestring, player->jointsMoving()? "moving\n\n" : "not moving\n\n");
            //Logger::logoutput(Logger::LOG_TEMP,log_parse_test,parse_test? "ok" : "error");
#endif
            
            if (world->playMode != PM_BeforeKickOff) {
                
                behave();
            }
            
            //add player model data to message
            player->buildMessage(message);
            
            //send message to server
            socket->send(message);
            
#ifdef _LOGGER
            Logger::logoutput(Logger::LOG_SENT_DATA, log_sent, message);
#endif
        }
    }
    
    return 0;
}

int Agent::behave() {
    
    int ret;
    
    if (actualLowSkill) {
        
        ret = actualLowSkill->calculate(NULL);
        if (ret == 0) {
            
            delete actualLowSkill;
            actualLowSkill = NULL;
        }
    }
    
    if (actualMiddleSkill) {
        
        ret = actualMiddleSkill->calculate((Skill **) & actualLowSkill);
        if (ret == 0) {
            
            delete actualMiddleSkill;
            actualMiddleSkill = NULL;
        }
    }
    
    if (actualHighSkill) {
        
        ret = actualHighSkill->calculate((Skill **) & actualMiddleSkill);
        if (ret == 0) {
            
            delete actualHighSkill;
            actualHighSkill = NULL;
        }
        
    } else {
        
        actualHighSkill = new HSAttack(world, player);
    }
    
    return 0;
}

void Agent::testSkill(char *skillName) {
    
    if (skillName != NULL) {
        
        actualHighSkill = new HSTest(world, player, skillName);
    }
}
