/*! \file Communication.cpp
 <pre>
 <b>File:</b> Communication.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/10/2008
 <b>Contents:</b> Communication module class id derived from Thread class, which makes it possible to run communication in separated thread.
 </pre>
 */

#include "Config.h"
#include "Communication.h"
#include "TCPSocket.h"
#include "UDPSocket.h"
#include "Logger.h"

#include <assert.h>
#include <iostream>

using namespace std;

extern Config *config;

Communication::Communication(char * serverIP, unsigned short serverPort) {
	
	serverAddrIn.sin_family			= AF_INET;
	serverAddrIn.sin_port			= htons(serverPort);
	serverAddrIn.sin_addr.s_addr	= inet_addr(serverIP);
    
    receivingMsg	= new char[NET_BUFFER_SIZE];
	receivingMsg[0] = '\0';
	
    pthread_mutex_init(&mutexSock, NULL);
    pthread_mutex_init(&mutexReceivingMsg, NULL);
    
	if (config->connectionType == CONN_TCP) {
        
		socket = new TCPSocket();
        
	} else if (config->connectionType == CONN_UDP) {
        
		socket = new UDPSocket();
	}
	
    connect();
}

Communication::~Communication() {
    
    close();
    pthread_mutex_destroy(&mutexSock);
    pthread_mutex_destroy(&mutexReceivingMsg);
    
	delete socket;
	delete[] receivingMsg;
}

bool Communication::connect() {
    
    pthread_mutex_lock(&mutexSock);
    bool result = socket->connect(&serverAddrIn);
    pthread_mutex_unlock(&mutexSock);
    
    return result;
}

bool Communication::close() {
    
    pthread_mutex_lock(&mutexSock);
    bool result = socket->close();
    pthread_mutex_unlock(&mutexSock);
    
    return result;
}

/*!
 * Wrapper for Thread::start method
 */
int Communication::startReceiving() {
    
#ifdef _LOGGER
    Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_starting);
#endif
    
	int rc = Thread::start();
	
#ifdef _LOGGER
    if (rc) Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_start_failed);
    else Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_started);
#endif
	
	if (rc) {
		cerr << "Failed to start communication" 
        << endl;
	}
	
	return rc;
}

/*!
 * Wrapper for Thread::stop method
 */
int Communication::stopReceiving() {
    
#ifdef _LOGGER
    Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_stopping);
#endif
    
	int rc = Thread::stop();
    
#ifdef _LOGGER
    if (rc) Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_stop_failed);
    else Logger::logoutput(Logger::LOG_COMMUNICATION,log_comm_stopped);
#endif
	
	if (rc) {
		cerr << "Failed to stop communication" 
        << endl;		
	} 
	
	return rc;
}

bool Communication::send(char * msg) {
    
	assert(msg != NULL);
    
    pthread_mutex_lock(&mutexSock);
    bool result = socket->send(msg);
    pthread_mutex_unlock(&mutexSock);
    
    return result;
}

bool Communication::receive() {
    
    pthread_mutex_lock(&mutexSock);
    pthread_mutex_lock(&mutexReceivingMsg);
    bool received = socket->receive(receivingMsg);
    pthread_mutex_unlock(&mutexReceivingMsg);
    pthread_mutex_unlock(&mutexSock);
    
    return received;
}

bool Communication::getMessage(char * msg) {
    
	assert(msg != NULL);
    
	pthread_mutex_lock(&mutexReceivingMsg);
	memcpy(msg, receivingMsg, strlen(receivingMsg) + 1);
	pthread_mutex_unlock(&mutexReceivingMsg);
    
    return true;
}

/*!
 * Communication thread routine.
 * This method is executed in separated thread.
 */
int Communication::run() {
    
    while (signal == THREADSIG_RUN) {
		
		receive();
    }
    
	return 0;
}
