/*! \file TSocket.cpp
 <pre>
 <b>File:</b> TSocket.cpp
 <b>Project:</b> Robocup 3D
 <b>Authors:</b> Hviezdna Jedenastka
 <b>Created:</b> 03/10/2008
 <b>Contents:</b> <TEXT>
 </pre>
 */

#include "TSocket.h"

#ifdef WIN32
#	include <winsock.h>
#else
#	define INVALID_SOCKET	-1
#	define SOCKET_ERROR		-1
#	include <netinet/in.h>
#	include <arpa/inet.h>
#endif

#include <iostream>
#include <assert.h>
#include <errno.h>
#include "Logger.h"

using namespace std;

TSocket::TSocket()
{
    startup();

    sockFD = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sockFD == INVALID_SOCKET) {
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_tcpsock_create_fail);
#endif
	cerr << "Failed to create TCP socket." << endl;
    }
}

bool TSocket::connect(const char *ip, unsigned short int port)
{
    assert(ip != NULL);
    assert(port != 0);

    server.sin_addr.s_addr = inet_addr(ip);
    server.sin_port = htons(port);
    server.sin_family = AF_INET;

    if (::connect(sockFD, (struct sockaddr *) &server, sizeof(server)) == 0) {
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_connected2, ip, port);
#endif
	return true;
    } else {
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_connect_fail2, ip, port);
#endif
	cerr << "Failed to connect to server: " << ip << ":" << port << "." << endl;
	return false;
    }
}

bool TSocket::send(const char *msg)
{
    //TODO_COMMENT
    char data[NET_BUFFER_SIZE + 1];
    int msgLength = (int) strlen(msg);
    int msgLength2 = htonl(msgLength);

    if (msgLength + sizeof(msgLength2) > NET_BUFFER_SIZE) {
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_out_msg_too_long, msgLength + sizeof(int), NET_BUFFER_SIZE);
#endif
	cerr << "Outgoing message is too long: " << msgLength + sizeof(msgLength2) << " max: " << NET_BUFFER_SIZE << endl;
    }

    memcpy(data, &msgLength2, sizeof(msgLength2));
    memcpy(data + sizeof(msgLength2), msg, msgLength);

    int sentBytes =::send(sockFD, data, sizeof(msgLength2) + msgLength, 0);

    switch (sentBytes) {
    case SOCKET_ERROR:
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_fail_sending);
#endif
	cerr << "Failed to send data." << endl;
	return false;

    default:
	if (sentBytes != (int) (sizeof(msgLength2) + msgLength)) {
#ifdef _LOGGER
	    Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_sending_error);
#endif
	    cerr << "Error sending data. Length of sent data does not match." << endl;
	    return false;
	}
    }

    return true;
}

bool TSocket::receive(char *buffer, unsigned int bufferSize)
{
    int receivedBytes;
    unsigned int messageSize;
    unsigned int bytesRead;

    bytesRead = 0;
    while (bytesRead < sizeof(unsigned int)) {
	receivedBytes =::recv(sockFD, buffer + bytesRead, sizeof(messageSize) - bytesRead, 0);
	//printf("%8x\n", buffer);
	if (receivedBytes < 0)
	    return false;

	bytesRead += receivedBytes;
    }

    /* convert from network byte order */
    messageSize = ntohl(*(unsigned int *) buffer);

    if (messageSize > bufferSize) {
	cerr << "Received data is too long: " << bufferSize << "." << endl;
	return false;
    }

    bytesRead = 0;
    while (bytesRead < messageSize) {
	receivedBytes =::recv(sockFD, buffer + bytesRead, messageSize - bytesRead, 0);
	if (receivedBytes < 0)
	    return false;

	bytesRead += receivedBytes;
    }

    buffer[bytesRead] = '\0';

    return true;
}

bool TSocket::select(unsigned int seconds, unsigned int microseconds)
{
    //TODO_COMMENT
    fd_set readset;
    struct timeval tv;
    int result;

    FD_ZERO(&readset);
    FD_SET(sockFD, &readset);
    tv.tv_sec = seconds;
    tv.tv_usec = microseconds;

    result =::select((int) sockFD + 1, &readset, NULL, NULL, &tv);
    switch (result) {
    case -1:
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_error_select, strerror(errno));
#endif
	cerr << "Error on select(): " << strerror(errno) << endl;
	return false;
    case 0:
	/* no data */
	return false;
    default:
	/* socket is ready */
	return true;
    }
}

bool TSocket::close()
{
    int result;
    result = closesocket(sockFD);

    if (result == SOCKET_ERROR) {
#ifdef _LOGGER
	Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_tcpsock_close_fail);
#endif
	cerr << "Failed to close TCP socket." << endl;
	return false;
    }

    cleanup();
#ifdef _LOGGER
    Logger::logoutput(Logger::LOG_COMMUNICATION, log_comm_tcpsock_close);
#endif
    return true;
}
